from .graph_utils import *
from .randomization import *
from .drawing import *
from .spectral import *
from .quantum import *

from scipy import linalg as la
from scipy.sparse.linalg import eigs
from sklearn.neighbors import NearestNeighbors
from scipy.stats import entropy

UNDIR_MATRIX_TYPES = ['adj', 'trans', 'lap', 'nlap', 'rwlap']
DIR_MATRIX_TYPES = ['adj', 'lap_out', 'lap_in']
MATRIX_TYPES = UNDIR_MATRIX_TYPES + DIR_MATRIX_TYPES
SUPPORTED_GRAPH_TYPES = [nx.Graph, nx.DiGraph]


def check_matrix_type(mode, is_directed):
    if mode not in MATRIX_TYPES:
        raise ValueError(f'Matrix type {mode} is not in allowed matrix types: {MATRIX_TYPES}')

    if is_directed and mode not in DIR_MATRIX_TYPES:
        raise ValueError(f'Matrix type {mode} is not allowed for directed networks.'
                         f'Supported options are: {DIR_MATRIX_TYPES}')

    if not is_directed and mode not in UNDIR_MATRIX_TYPES:
        raise ValueError(f'Matrix type {mode} is not allowed for undirected networks.'
                         f'Supported options are: {UNDIR_MATRIX_TYPES}')


def check_adjacency(a):
    if a.shape[0] != a.shape[1]:
        raise Exception('Non-square adjacency matrix!')


def check_directed(directed, real_world):
    if real_world:
        if int(directed) not in [0, 1]:
            raise Exception('Fractional direction is not valid for a real network')
    elif directed < 0 or directed > 1:
        raise Exception('Wrong "directed" parameter value:', directed)


def check_weights_and_directions(a, weighted, directed):
    is_directed = not np.allclose(a.data, a.T.data)
    is_weighted = not np.allclose(a.data, a.astype(bool).astype(int).data)

    symm_text = 'asymmetric' if is_directed else 'symmetric'
    if is_directed != bool(directed):
        raise Exception(f'Error in network construction: "directed" set to {directed},'
                        f' but the adjacency matrix is {symm_text}')

    w_text = 'weighted' if is_weighted else 'not weighted'
    if is_weighted != bool(weighted):
        raise Exception(f'Error in network construction: "weighted" set to {weighted},'
                        f' but the adjacency matrix {w_text}')


def select_construction_pipeline(a, graph):
    # TODO: add partial directions
    if a is None:
        if graph is None:
            raise ValueError('Either "adj" or "graph" argument must be non-empty')
        else:
            if not np.any([isinstance(graph, gtype) for gtype in SUPPORTED_GRAPH_TYPES]):
                raise TypeError(f'graph should have one of supported graph types: {SUPPORTED_GRAPH_TYPES}')
            else:
                pipeline = 'graph'

    else:
        if graph is None:
            pipeline = 'adj'
        else:
            raise ValueError('Either "adj" or "graph" should be given, not both')

    return pipeline


class Network:
    """
    An object for network analysis with the focus on spectral graph theory
    """

    def __init__(self,
                 adj=None,
                 graph=None,
                 preprocessing='giant_cc',
                 name='',
                 pos=None,
                 verbose=False,
                 create_nx_graph=True,
                 node_attrs=None,
                 **network_args):

        self.name = name
        self.verbose = verbose
        self.network_params = network_args
        self.create_nx_graph = create_nx_graph

        self.init_method = select_construction_pipeline(adj, graph)

        self.directed = network_args.get('directed')
        if self.directed is None:
            if self.init_method == 'adj':
                self.directed = not np.allclose(adj.data, adj.T.data)
            elif self.init_method == 'graph':
                self.directed = nx.is_directed(graph)

        self.weighted = network_args.get('weighted')
        if self.weighted is None:
            if self.init_method == 'adj':
                self.weighted = not np.allclose(adj.data, adj.astype(bool).astype(int).data)
            elif self.init_method == 'graph':
                self.weighted = nx.is_weighted(graph)

        self.real_world = network_args.get('real_world')
        if self.real_world is None:
            self.real_world = True

        check_directed(self.directed, self.real_world)

        # set empty attributes for different matrix and data types
        valid_mtypes = DIR_MATRIX_TYPES if self.directed else UNDIR_MATRIX_TYPES
        for mt in valid_mtypes:
            setattr(self, mt, None)
            setattr(self, mt + '_spectrum', None)
            setattr(self, mt + '_eigenvectors', None)
            setattr(self, mt + '_zvalues', None)
            setattr(self, mt + '_ipr', None)

        # initialize adjacency matrix and (probably) associated graph
        if self.init_method == 'adj':
            # initialize Network object from sparse matrix
            self._preprocess_adj_and_data(a=adj,
                                          pos=pos,
                                          node_attrs=node_attrs,
                                          preprocessing=preprocessing,
                                          create_graph=create_nx_graph)

        if self.init_method == 'graph':
            # initialize Network object from NetworkX graph or digraph
            self._preprocess_graph_and_data(graph=graph,
                                            pos=pos,
                                            node_attrs=node_attrs,
                                            preprocessing=preprocessing)

        # each network object has out- and in-degree sequences from its initialization
        self.get_node_degrees()

        # TODO: remove lem_emb
        self.lem_emb = None

    def _preprocess_graph_and_data(self,
                                   graph=None,
                                   preprocessing=None,
                                   pos=None,
                                   node_attrs=None):

        if preprocessing is None:
            if self.verbose:
                print('No preprocessing specified, this may lead to unexpected errors in graph connectivity!')
            fgraph = remove_selfloops_from_graph(graph)

        elif preprocessing == 'remove_isolates':
            fgraph = remove_isolates_and_selfloops_from_graph(graph)

        elif preprocessing == 'giant_cc':
            g_ = remove_selfloops_from_graph(graph)
            fgraph = get_giant_cc_from_graph(g_)
            self.n_cc = 1

        elif preprocessing == 'giant_scc':
            g_ = remove_selfloops_from_graph(graph)
            fgraph = get_giant_scc_from_graph(g_)
            self.n_cc = 1
            self.n_scc = 1

        else:
            raise ValueError('Wrong preprocessing type!')

        lost_nodes = graph.number_of_nodes() - fgraph.number_of_nodes()
        lost_edges = graph.number_of_edges() - fgraph.number_of_edges()
        if lost_nodes + lost_edges != 0 and self.verbose:
            print(f'{lost_nodes} nodes and {lost_edges} edges removed ')

        # add node positions if provided
        if pos is not None:
            self.pos = {node: pos[node] for node in graph.nodes() if node in fgraph.nodes()}
        else:
            self.pos = None

        if node_attrs is not None:
            self.node_attrs = {node: node_attrs[node] for node in graph.nodes() if node in fgraph.nodes()}
        else:
            self.node_attrs = None

        self.graph = fgraph
        self.adj = nx.adjacency_matrix(fgraph)
        self.n = nx.number_of_nodes(self.graph)

    def _preprocess_adj_and_data(self,
                                 a=None,
                                 preprocessing=None,
                                 pos=None,
                                 node_attrs=None,
                                 create_graph=True):

        # if NetworkX graph should be created, we revert to graph-based initialization for simplicity
        if create_graph:
            gtype = nx.DiGraph if self.directed else nx.Graph
            graph = nx.from_scipy_sparse_array(a, create_using=gtype)
            self._preprocess_graph_and_data(graph=graph,
                                            pos=pos,
                                            node_attrs=node_attrs,
                                            preprocessing=preprocessing)

            return

        if preprocessing is None:
            if self.verbose:
                print('No preprocessing specified, this may lead to unexpected errors in graph connectivity!')
            fadj = remove_selfloops_from_adj(a)
            nodes_range = range(fadj.shape[0])
            node_mapping = dict(zip(nodes_range, nodes_range))  # no nodes have been deleted

        elif preprocessing == 'remove_isolates':
            a_ = remove_selfloops_from_adj(a)
            fadj, node_mapping = remove_isolates_from_adj(a_)

        elif preprocessing == 'giant_cc':
            a_ = remove_selfloops_from_adj(a)
            fadj, node_mapping = get_giant_cc_from_adj(a_)
            self.n_cc = 1

        elif preprocessing == 'giant_scc':
            a_ = remove_selfloops_from_adj(a)
            fadj, node_mapping = get_giant_scc_from_adj(a_)
            self.n_cc = 1
            self.n_scc = 1

        else:
            raise ValueError('Wrong preprocessing type!')

        lost_nodes = a.shape[0] - fadj.shape[0]
        lost_edges = a.nnz - fadj.nnz
        if not self.directed:
            lost_edges = lost_edges // 2

        if lost_nodes + lost_edges != 0 and self.verbose:
            print(f'{lost_nodes} nodes and {lost_edges} edges removed')

        # add node positions if provided
        if pos is not None:
            self.pos = {node: pos[node] for node in range(a.shape[0]) if node in node_mapping}
        else:
            self.pos = None

        if node_attrs is not None:
            self.node_attrs = {node: node_attrs[node] for node in range(a.shape[0]) if node in node_mapping}
        else:
            self.node_attrs = None

        self.graph = None
        self.adj = fadj
        self._init_to_final_node_mapping = node_mapping
        self.n = self.adj.shape[0]

    def is_connected(self):
        ccs = list(get_ccs_from_adj(self.adj))
        return len(ccs) == 1

    def randomize(self, rmode='shuffle'):
        # TODO: update routines
        if rmode == 'graph_iom':
            if self.directed:
                g = nx.DiGraph(self.graph)
            else:
                g = nx.Graph(self.graph)

            new_graph = random_rewiring_IOM_preserving(g, r=10)
            rand_adj = nx.adjacency_matrix(new_graph)

        elif rmode == 'adj_iom':
            rand_adj = adj_random_rewiring_iom_preserving(self.adj,
                                                          is_weighted=self.weighted,
                                                          r=10)

        elif rmode == 'complete':
            rand_adj = random_rewiring_complete_graph(self.adj)

        elif rmode == 'shuffle':
            rand_adj = random_rewiring_dense_graph(self.adj)

        else:
            raise ValueError('Unknown randomization method')

        rand_net = Network(sp.csr_matrix(rand_adj),
                           self.network_params,
                           name=self.name + f' {rmode} rand',
                           pos=self.pos,
                           real_world=False,
                           verbose=False)

        return rand_net

    def get_node_degrees(self):
        # convert sparse matrix to 0-1 format and sum over specific axis
        self.outdeg = np.array(self.adj.astype(bool).astype(int).sum(axis=0)).ravel()
        self.indeg = np.array(self.adj.astype(bool).astype(int).sum(axis=1)).ravel()
        self.deg = np.array((self.adj + self.adj.T).astype(bool).astype(int).sum(axis=1)).ravel()

        min_out = min(self.outdeg)
        min_in = min(self.indeg)
        max_out = max(self.outdeg)
        max_in = max(self.indeg)

        if max_out != min_out:
            self.scaled_outdeg = np.array([1.0 * (deg - min_out) / (max_out - min_out) for deg in self.outdeg])
        else:
            self.scaled_outdeg = np.ones(len(self.outdeg))

        if min_in != max_in:
            self.scaled_indeg = np.array([1.0 * (deg - min_in) / (max_in - min_in) for deg in self.indeg])
        else:
            self.scaled_outdeg = np.ones(len(self.indeg))

    def get_degree_distr(self, mode='all'):
        if mode == 'all':
            deg = self.deg
        elif mode == 'out':
            deg = self.outdeg
        elif mode == 'in':
            deg = self.indeg
        else:
            raise ValueError('Wrong mode for degree distribution.')

        hist, bins = np.histogram(deg, bins=max(deg) - min(deg), density=True)
        return hist

    def get_matrix(self, mode):
        check_matrix_type(mode, self.directed)
        matrix = getattr(self, mode)
        if matrix is None:
            if mode == 'lap' or mode == 'lap_out':
                matrix = get_laplacian(self.adj)
            elif mode == 'nlap':
                matrix = get_norm_laplacian(self.adj)
            elif mode == 'rwlap':
                matrix = get_rw_laplacian(self.adj)
            elif mode == 'trans':
                matrix = get_trans_matrix(self.adj)
            elif mode == 'adj':
                matrix = self.adj
            else:
                raise Exception(f'Wrong matrix type: {mode}')

            setattr(self, mode, matrix)

        matrix = getattr(self, mode)
        return matrix

    def get_spectrum(self, mode):
        check_matrix_type(mode, self.directed)
        spectrum = getattr(self, mode + '_spectrum')
        if spectrum is None:
            self.diagonalize(mode=mode)
            spectrum = getattr(self, mode + '_spectrum')

        return spectrum

    def get_eigenvectors(self, mode):
        check_matrix_type(mode, self.directed)
        eigenvectors = getattr(self, mode + '_eigenvectors')
        if eigenvectors is None:
            self.diagonalize(mode=mode)
            eigenvectors = getattr(self, mode + '_eigenvectors')

        return eigenvectors

    def get_ipr(self, mode):
        check_matrix_type(mode, self.directed)
        ipr = getattr(self, mode + '_ipr')
        if ipr is None:
            self.calculate_ipr(mode=mode)
            ipr = getattr(self, mode + '_ipr')

        return ipr

    def get_z_values(self, mode):
        check_matrix_type(mode, self.directed)
        zvals = getattr(self, mode + '_zvalues')
        if zvals is None:
            self.calculate_z_values(mode=mode)
            zvals = getattr(self, mode + '_zvalues')

        return zvals

    def partial_diagonalize(self, spectrum_params):
        '''
        noise = self.spectrum_params['noise']
        neigs = self.spectrum_params['neigs']
        if noise == 0:
            R = np.zeros((n, n))
        else:
            R = np.multiply(matrix.A, np.random.normal(loc=0.0, scale=noise, size=(n, n)))
        '''
        raise Exception('this method is under construction')

    def diagonalize(self, mode='lap', verbose=None):
        if verbose is None:
            verbose = self.verbose

        check_matrix_type(mode, self.directed)
        if verbose:
            print('Preparing for diagonalizing...')

        outdeg = self.outdeg
        indeg = self.indeg
        deg = self.deg

        A = self.adj.astype(float)
        n = self.n

        if n != np.count_nonzero(outdeg) and verbose:
            print(n - np.count_nonzero(outdeg), 'nodes without out-edges')
        if n != np.count_nonzero(indeg) and verbose:
            print(n - np.count_nonzero(indeg), 'nodes without in-edges')

        nz = np.count_nonzero(deg)
        if nz != n and self.n_cc == 1:
            print('Graph has', str(n - nz), 'isolated nodes!')
            raise Exception('Graph is not connected!')

        if not self.weighted and not self.directed and not np.allclose(outdeg, indeg):
            raise Exception('out- and in- degrees do not coincide in boolean')

        matrix = self.get_matrix(mode)

        if verbose:
            print('Performing diagonalization...')

        matrix_is_symmetric = np.allclose(matrix.data, matrix.T.data)
        if matrix_is_symmetric:
            raw_eigs, right_eigvecs = la.eigh(matrix.A)
        else:
            raw_eigs, right_eigvecs = la.eig(matrix.A, right=True)

        raw_eigs = np.around(raw_eigs, decimals=12)
        sorted_eigs = np.sort(raw_eigs)

        if 'lap' in mode:
            n_comp = len(raw_eigs[np.abs(raw_eigs) == 0])
            if n_comp != 1 and not self.weighted and self.n_cc == 1:
                print('eigenvalues:', sorted_eigs)
                raise Exception('Graph has %d components!' % n_comp)

        setattr(self, mode, matrix)

        if np.allclose(np.imag(sorted_eigs), np.zeros(len(sorted_eigs)), atol=1e-12):
            sorted_eigs = np.real(sorted_eigs)
        else:
            if not self.directed:
                raise ValueError('Complex eigenvalues found in non-directed network!')

        setattr(self, mode + '_spectrum', sorted_eigs)

        sorted_eigenvectors = right_eigvecs[np.ix_(range(len(sorted_eigs)), np.argsort(raw_eigs))]
        if np.allclose(np.imag(sorted_eigenvectors), np.zeros(sorted_eigenvectors.shape), atol=1e-8):
            sorted_eigenvectors = np.real(sorted_eigenvectors)
        else:
            if not self.directed:
                raise ValueError('Complex eigenvectors found in non-directed network!')

        setattr(self, mode + '_eigenvectors', sorted_eigenvectors)

        if verbose:
            print('Diagonalizing finished')

    # TODO: add Gromov hyperbolicity

    def calculate_z_values(self, mode='lap'):
        spectrum = self.get_spectrum(mode)
        seigs = sorted(list(set(spectrum)), key=np.abs)
        if len(seigs) != len(spectrum) and self.verbose:
            print('WARNING:', len(spectrum) - len(seigs), 'repeated eigenvalues discarded')

        if self.verbose:
            print('Computing nearest neighbours...')

        X = np.array([[np.real(x), np.imag(x)] for x in seigs])
        nbrs = NearestNeighbors(n_neighbors=3, algorithm='ball_tree').fit(X)
        distances, indices = nbrs.kneighbors(X)
        nnbs = [seigs[x] for x in indices[:, 1]]
        nnnbs = [seigs[x] for x in indices[:, 2]]
        '''
        nndist = np.array([(nnbs[i] - eigs[i]) for i in range(len(eigs))])
        nnndist = np.array([(nnnbs[i] - eigs[i]) for i in range(len(eigs))])
        '''
        zlist = np.array([(nnbs[i] - seigs[i]) / (nnnbs[i] - seigs[i]) for i in range(len(seigs))])
        zdict = dict(zip(seigs, zlist))

        setattr(self, mode + '_zvalues', zdict)

    def calculate_ipr(self, mode='adj'):
        eigenvectors = self.get_eigenvectors(mode)
        nvecs = eigenvectors.shape[1]
        ipr = np.zeros(nvecs)
        eig_entropy = np.zeros(nvecs)

        for i in range(nvecs):
            ipr[i] = sum([np.abs(v) ** 4 for v in eigenvectors[:, i]])
            # entropy[i] = -np.log(ipr[i]) # erdos entropy (deprecated)
            eig_entropy[i] = entropy(np.array([np.abs(v) ** 2 for v in eigenvectors[:, i]]))

        setattr(self, mode + '_ipr', ipr)
        # self.eigenvector_entropy = eig_entropy / np.log(self.n)

    def _get_lap_spectrum(self, norm=False):
        if not self.directed:
            if norm:
                spectrum = self.get_spectrum('nlap') # could be rwlap as well
            else:
                spectrum = self.get_spectrum('lap')
        else:
            if norm:
                raise NotImplementedError('Normalized Laplacian not implemented for directed networks')
            else:
                spectrum = self.get_spectrum('lap_out')

        return spectrum

    def calculate_thermodynamic_entropy(self, tlist, verbose=False, norm=False):
        spectrum = self._get_lap_spectrum(norm=norm)
        res = [spectral_entropy(spectrum, t, verbose=verbose) for t in tlist]
        return res

    def calculate_free_entropy(self, tlist, norm=False):
        spectrum = self._get_lap_spectrum(norm=norm)
        res = [free_entropy(spectrum, t) for t in tlist]
        return res

    def calculate_q_entropy(self, q, tlist, norm=False):
        spectrum = self._get_lap_spectrum(norm=norm)
        res = [q_entropy(spectrum, t, q=q) for t in tlist]
        return res

    def calculate_estrada_communicability(self):
        adj_spectrum = self.get_spectrum('adj')
        self.estrada_communicability = sum([np.exp(e) for e in adj_spectrum])
        return self.estrada_communicability

    def get_estrada_bipartivity_index(self):
        adj_spectrum = self.get_spectrum('adj')
        esi1 = sum([np.exp(-e) for e in adj_spectrum])
        esi2 = sum([np.exp(e) for e in adj_spectrum])
        self.estrada_bipartivity = esi1 / esi2
        return self.estrada_bipartivity

    def localization_signatures(self, mode='lap'):
        zvals = self.get_z_values(mode)

        mean_cos_phi = np.mean(np.array([np.cos(np.angle(x)) for x in zvals]))
        rvals = [1. / (np.abs(z)) ** 2 for z in zvals]
        mean_inv_r_sq = np.mean(np.array(rvals))

        if self.verbose:
            print('mean cos phi complex:', mean_cos_phi)
            print('mean 1/r^2 real:', mean_inv_r_sq)

        return mean_inv_r_sq, mean_cos_phi

    def construct_lem_embedding(self, dim):
        if self.directed:
            raise Exception('LEM embedding is not implemented for directed graphs')

        A = self.adj
        A = A.asfptype()
        print('Performing spectral decomposition...')
        K = A.shape[0]
        NL = get_norm_laplacian(A)
        DH = get_inv_sqrt_diag_matrix(A)

        start_v = np.ones(K)
        eigvals, eigvecs = eigs(NL, k=dim + 1, which='LR', v0=start_v, maxiter=K * 1000)
        eigvals = np.asarray([np.round(np.real(x), 6) for x in eigvals])

        if np.count_nonzero(eigvals == 1.0) > 1:
            raise Exception('Error while LEM embedding construction: graph is not connected!')
        else:
            vecs = eigvecs.T[1:]
            vec_norms = np.array([np.real(sum([x * x for x in v])) for v in vecs])
            vecs = vecs / vec_norms[:, np.newaxis]
            # explanation: https://jlmelville.github.io/smallvis/spectral.html
            vecs = DH.dot(sp.csr_array(vecs, dtype=float))

            self.lem_emb = vecs



from pydrive2.auth import GoogleAuth


def desktop_auth(secret_path):
    gauth = GoogleAuth()
    GoogleAuth.DEFAULT_SETTINGS['client_config_file'] = secret_path
    # Create local webserver and auto handles authentication.
    gauth.LocalWebserverAuth()
    return gauth


def google_colab_auth():
    from google.colab import auth
    from oauth2client.client import GoogleCredentials
    auth.authenticate_user()
    gauth = GoogleAuth()
    gauth.credentials = GoogleCredentials.get_application_default()
    return gauth


from os.path import join

from pydrive2.drive import GoogleDrive
import warnings
import wget
import gdown
import pandas as pd
import shutil

from .gdrive_utils import *
from ..utils.output import *

def retrieve_relevant_ids(folder, name_part, whitelist=[], extensions=['.csv', '.xlsx']):
    return_code = True
    folder_page = client.get(folder)

    if folder_page.status_code != 200:
        return False, None

    gdrive_file, id_name_type_iter = parse_google_drive_file(
        folder,
        folder_page.text,
    )

    relevant = []
    if len(list(id_name_type_iter)) > MAX_NUMBER_FILES:
        raise MemoryError(
            f'The folder {folder} has {len(list(id_name_type_iter))} elements while max allowed number of files is {MAX_NUMBER_FILES}')

    for child_id, child_name, child_type in id_name_type_iter:
        if child_type != folder_type:
            if (name_part in child_name or child_name in whitelist) and os.path.splitext(child_name)[1] in extensions:
                relevant.append((child_id, child_name))

        else:
            return_code, rel_sublist = retrieve_relevant_ids(folders_url + child_id, name_part)
            if not return_code:
                print(f'recursive search broke on folder {child_id}')
                break
            relevant.extend(rel_sublist)

    return return_code, relevant


def download_part_of_folder(
        output,  # path for downloaded data
        folder,  # share link to google drive folder
        key='',  # part of filename to search for
        whitelist=[],  # list of filenames to be downloaded regardless of their names
        extensions=['.csv', '.xlsx', '.npz'],  # allowed file extensions
        via_pydrive=False,  # pydrive requires authorization, but can download a big number of files,
        gauth=None,
        maxfiles=None):

    with Capturing() as load_log:
        if via_pydrive:
            if gauth is None:
                raise ValueError('To use pydrive, you need to authenticate using one of the functions'
                                 ' in driada.gdrive.auth')
            drive = GoogleDrive(gauth)

            rel = []
            fid = id_from_link(folder)
            file_list = drive.ListFile({'q': f"'{fid}' in parents and trashed=false"}).GetList()
            if maxfiles is not None:
                file_list = file_list[:maxfiles]

            for f in file_list:
                if key in f['title']:
                    # print('title: %s, id: %s' % (f['title'],f['id']))
                    os.makedirs(output, exist_ok=True)
                    f.GetContentFile(join(output, f['title']))
                    rel.append((f['id'], f['title']))

            return_code = True

        else:
            return_code, rel = retrieve_relevant_ids(folder,
                                                     key,
                                                     whitelist=whitelist,
                                                     extensions=extensions)

            if return_code:
                for i, pair in enumerate(rel):
                    idx, name = rel[i]
                    gdown.download(id=idx, output=os.path.join(output, name))

            else:
                raise FileNotFoundError('Error in downloading procedure!')

        return return_code, rel, load_log


def download_gdrive_data(data_router,
                         expname,
                         whitelist=['Timing.xlsx'],
                         via_pydrive=False,
                         data_pieces=None,
                         tdir='DRIADA data',
                         gauth=None):

    with Capturing() as load_log:
        print('-------------------------------------------------------------')
        print(f'Extracting data for {expname} from Google Drive')
        print('-------------------------------------------------------------')

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=DeprecationWarning)

            success = False
            available_exp = data_router['Эксперимент'].values
            if expname not in available_exp:
                print(f'{expname} not found in available experiments: {available_exp}')
                return success, load_log

            row = data_router[data_router['Эксперимент'] == expname]
            links = dict(zip(row.columns, row.values[0]))

            os.makedirs(join(tdir, expname), exist_ok=True)
            if data_pieces is None:
                data_pieces = [d for d in list(data_router.columns.values) if d not in ['Эксперимент', 'Краткое описание', 'Video', 'Aligned data', 'Computation results']]

            for key in data_pieces:
                if 'http' in links[key]:
                    print(f'Loading data: {key}...')
                    ddir = join(tdir, expname, key)
                    os.makedirs(ddir, exist_ok=True)
                    # gdown.download_folder(url = links[key], output = dir, quiet=False)
                    return_code, rel, folder_log = download_part_of_folder(ddir,
                                                                           links[key],
                                                                           key=expname,
                                                                           whitelist=whitelist,
                                                                           via_pydrive=via_pydrive,
                                                                           gauth=gauth)

                    load_log.extend(folder_log)

                    if len(rel) == 0:
                        os.rmdir(ddir)
                        print('No relevant data found')
                    else:
                        loaded_names = [r[1] for r in rel]
                        for n in loaded_names:
                            print(n)

                        success = True

                    print('--------------------------')

            return success, load_log


def initialize_iabs_router(root='\\content'):
    router_name = 'IABS data router.xlsx'
    router_path = join(root, router_name)
    os.makedirs(root, exist_ok=True)
    if router_name in os.listdir(root):
        os.remove(router_path)

    global_data_table_url = 'https://docs.google.com/spreadsheets/d/130DDFAoAbmm0jcKLBF6xsWsQLDr2Zsj4cPuOYivXoM8/export?format=xlsx'
    wget.download(global_data_table_url, out=router_path)

    data_router = pd.read_excel(router_path)
    #data_router.fillna(method='ffill', inplace=True)
    data_router = data_router.replace("", None).ffill()

    data_pieces = [d for d in list(data_router.columns.values) if d not in ['Эксперимент',
                                                                            'Краткое описание',
                                                                            'Video',
                                                                            'Aligned data',
                                                                            'Computation results']
                   ]
    return data_router, data_pieces



import requests
import regex
import os
from bs4 import BeautifulSoup
from itertools import islice
import json

class GoogleDriveFile(object):
    """Represent Google Drive file objects structure.

    Attributes
    ----------
    id: str
        Unique id, used to build the download URL.
    name: str
        Actual name, used as file name.
    type: str
        MIME type, or application/vnd.google-apps.folder if it is a folder
    children: List[GoogleDriveFile]
        If it is a directory, it contains the folder files/directories

    """

    def __init__(self, id, name, type, children=None):
        self.id = id
        self.name = name
        self.type = type
        self.children = children if children is not None else []

    def is_folder(self):
        return self.type == folder_type

    def __repr__(self):
        template = "(id={id}, name={name}, type={type}, children={children})"
        return "GoogleDriveFile" + template.format(
            id=self.id,
            name=self.name,
            type=self.type,
            children=self.children,
        )


def parse_google_drive_file(folder, content, use_cookies=True):
    """Extracts information about the current page file and its children

    Parameters
    ----------
    folder: str
        URL of the Google Drive folder.
        Must be of the format 'https://drive.google.com/drive/folders/{url}'.
    content: str
        Google Drive's raw string

    Returns
    -------
    gdrive_file: GoogleDriveFile
        Current GoogleDriveFile, with empty children
    id_name_type_iter: Iterator
        Tuple iterator of each children id, name, type
    """
    folder_soup = BeautifulSoup(content, features="html.parser")

    if not use_cookies:
        client.cookies.clear()

    # finds the script tag with window['_DRIVE_ivd']
    encoded_data = None
    for script in folder_soup.select("script"):
        inner_html = script.decode_contents()

        if "_DRIVE_ivd" in inner_html:
            # first js string is _DRIVE_ivd, the second one is the encoded arr
            regex_iter = string_regex.finditer(inner_html)
            # get the second elem in the iter
            try:
                encoded_data = next(islice(regex_iter, 1, None)).group(1)
            except StopIteration:
                raise RuntimeError(
                    "Couldn't find the folder encoded JS string"
                )
            break

    if encoded_data is None:
        raise RuntimeError(
            "Cannot retrieve the folder information from the link. "
            "You may need to change the permission to "
            "'Anyone with the link'."
        )

    # decodes the array and evaluates it as a python array
    decoded = encoded_data.encode("utf-8").decode("unicode_escape")
    folder_arr = json.loads(decoded)

    folder_contents = [] if folder_arr[0] is None else folder_arr[0]

    gdrive_file = GoogleDriveFile(
        id=folder.split("/")[-1],
        name=" - ".join(folder_soup.title.contents[0].split(" - ")[:-1]),
        type=folder_type,
    )

    id_name_type_iter = [
        (e[0], e[2].encode("raw_unicode_escape").decode("utf-8"), e[3])
        for e in folder_contents
    ]

    return gdrive_file, id_name_type_iter


def download_and_parse_google_drive_link(
        folder,
        quiet=False,
        use_cookies=True,
        remaining_ok=False,
        name_part=''
):
    """Get folder structure of Google Drive folder URL.

    Parameters
    ----------
    folder: str
        URL of the Google Drive folder.
        Must be of the format 'https://drive.google.com/drive/folders/{url}'.
    quiet: bool, optional
        Suppress terminal output.
    use_cookies: bool, optional
        Flag to use cookies. Default is True.
    remaining_ok: bool, optional
        Flag that ensures that is ok to let some file to not be downloaded,
        since there is a limitation of how many items gdown can download,
        default is False.

    Returns
    -------
    return_code: bool
        Returns False if the download completed unsuccessfully.
        May be due to invalid URLs, permission errors, rate limits, etc.
    gdrive_file: GoogleDriveFile
        Returns the folder structure of the Google Drive folder.
    """
    return_code = True

    folder_page = client.get(folder)

    if folder_page.status_code != 200:
        return False, None

    gdrive_file, id_name_type_iter = parse_google_drive_file(
        folder,
        folder_page.text,
    )

    for child_id, child_name, child_type in id_name_type_iter:
        if name_part in child_name:
            if child_type != folder_type:
                if not quiet:
                    print(
                        "Processing file",
                        child_id,
                        child_name,
                    )
                gdrive_file.children.append(
                    GoogleDriveFile(
                        id=child_id,
                        name=child_name,
                        type=child_type,
                    )
                )
                if not return_code:
                    return return_code, None
                continue

            if not quiet:
                print(
                    "Retrieving folder",
                    child_id,
                    child_name,
                )
            return_code, child = download_and_parse_google_drive_link(
                folders_url + child_id,
                use_cookies=use_cookies,
                quiet=quiet,
            )
            if not return_code:
                return return_code, None
            gdrive_file.children.append(child)

    has_at_least_max_files = len(gdrive_file.children) == MAX_NUMBER_FILES
    if not remaining_ok and has_at_least_max_files:
        err_msg = " ".join(
            [
                "The gdrive folder with url: {url}".format(url=folder),
                "has at least {max} files,".format(max=MAX_NUMBER_FILES),
                "gdrive can't download more than this limit,",
                "if you are ok with this,",
                "please run again with --remaining-ok flag.",
            ]
        )
        raise RuntimeError(err_msg)
    return return_code, gdrive_file


def id_from_link(link):
    if 'http' not in link:
        raise ValueError('Wrong link format')

    if 'id=' in link:
        return link.split('id=')[-1].split('&')[0]
    else:
        return link.split('folders/')[-1].split('?')[0]


folders_url = "https://drive.google.com/drive/folders/"
files_url = "https://drive.google.com/uc?id="
folder_type = "application/vnd.google-apps.folder"

string_regex = regex.compile(r"'((?:[^'\\]|\\.)*)'")
MAX_NUMBER_FILES = 50

client = requests.session()





import copy
import os.path
import numpy as np
import pickle

from .exp_base import Experiment
from ..information.info_base import TimeSeries
from ..utils.naming import construct_session_name
from ..utils.output import show_output
from .neuron import DEFAULT_FPS, DEFAULT_T_OFF, DEFAULT_T_RISE
from ..gdrive.download import download_gdrive_data, initialize_iabs_router


def load_exp_from_aligned_data(data_source,
                               exp_params,
                               data,
                               force_continuous=[],
                               bad_frames=[],
                               static_features=None,
                               verbose=True,
                               reconstruct_spikes='wavelet'):

    expname = construct_session_name(data_source, exp_params)
    adata = copy.deepcopy(data)
    key_mapping = {key.lower(): key for key in adata.keys()}

    if verbose:
        print(f'Building experiment {expname}...')

    if 'calcium' in key_mapping:
        calcium = adata.pop(key_mapping['calcium'])
    else:
        raise ValueError('No calcium data found!')

    spikes = None
    if 'spikes' in key_mapping:
        spikes = adata.pop(key_mapping['spikes'])

    dyn_features = adata.copy()

    def is_garbage(vals):
        return len(set(vals)) == 1 or np.sum(np.isnan(vals)).astype(int) == len(vals)

    if len(force_continuous) != 0:
        feat_is_continuous = {f: f in force_continuous for f in dyn_features.keys()}
        filt_dyn_features = {f: TimeSeries(vals, discrete=not feat_is_continuous[f]) for f, vals in dyn_features.items() if not is_garbage(vals)}
    else:
        filt_dyn_features = {f: TimeSeries(vals) for f, vals in dyn_features.items() if not is_garbage(vals)}

    if verbose:
        print('behaviour variables:')
        print()
        for f, ts in filt_dyn_features.items():
            print(f"'{f}'", 'discrete' if ts.discrete else 'continuous')

    # check for constant features
    constfeats = set(dyn_features.keys() - set(filt_dyn_features.keys()))

    if len(constfeats) != 0 and verbose:
        print(f'features {constfeats} dropped as constant or empty')

    auto_continuous = [fn for fn, ts in filt_dyn_features.items() if not ts.discrete]
    if verbose:
        print(f'features {auto_continuous} automatically determined as continuous')
        print()

    if len(force_continuous) != 0:
        if set(auto_continuous) != (set(force_continuous) & set(dyn_features.keys())):
            print('Warning: auto determined continuous features do not coincide with force_continuous list! Automatic labelling will be overridden')
            for fn, ts in filt_dyn_features.items():
                if len(set(ts.data)) > 2 and not(fn in force_continuous):
                    filt_dyn_features[fn] = TimeSeries(ts.data.astype(bool).astype(int), discrete=True)
                    if verbose:
                        print(f'feature {fn} converted to integer')

    signature = f'Exp {expname}'

    # set default static experiment features if not provided
    default_static_features = {'t_rise_sec': DEFAULT_T_RISE,
                               't_off_sec': DEFAULT_T_OFF,
                               'fps': DEFAULT_FPS}

    if static_features is None:
        static_features = dict()
    for sf in default_static_features.keys():
        if sf not in static_features:
            static_features.update({sf: default_static_features[sf]})

    exp = Experiment(signature,
                     calcium,
                     spikes,
                     exp_params,
                     static_features,
                     filt_dyn_features,
                     reconstruct_spikes=reconstruct_spikes,
                     bad_frames_mask=np.array([True if _ in bad_frames else False for _ in range(calcium.shape[1])])
                     )


    return exp


# TODO: add support for g-cloud auth from desktop (for pydrive) https://cloud.google.com/docs/authentication/application-default-credentials
def load_experiment(data_source,
                    exp_params,
                    force_rebuild=False,
                    force_reload=False,
                    via_pydrive=True,
                    gauth=None,
                    root='DRIADA data',
                    force_continuous=[],
                    bad_frames=[],
                    static_features=None,
                    reconstruct_spikes='wavelet',
                    verbose=True):

    os.makedirs(root, exist_ok=True)
    if not os.path.isdir(root):
        raise ValueError('Root must be a folder!')

    expname = construct_session_name(data_source, exp_params)
    exppath = os.path.join(root, expname, f'Exp {expname}.pickle')
    if os.path.exists(exppath) and not force_rebuild and not force_reload:
        Exp = load_exp_from_pickle(exppath, verbose=verbose)
        return Exp, None

    else:
        if data_source == 'IABS':
            syn_data_name = os.path.join(root,
                                         expname,
                                         'Aligned data',
                                         f'{expname} syn data.npz')

            data_exists = os.path.exists(syn_data_name)
            if force_reload or not data_exists:
                data_router, data_pieces = initialize_iabs_router(root=root)
                success, load_log = download_gdrive_data(data_router,
                                                         expname,
                                                         data_pieces=['Aligned data'],
                                                         via_pydrive=via_pydrive,
                                                         tdir=root,
                                                         gauth=gauth)

                if not success:
                    print('===========   BEGINNING OF LOADING LOG:   ============')
                    show_output(load_log)
                    print('===========   END OF LOADING LOG:   ============')
                    raise FileNotFoundError(f'Cannot download {expname}, see loading log above')

            else:
                load_log = None

            aligned_data = dict(np.load(syn_data_name))
            Exp = load_exp_from_aligned_data(data_source,
                                             exp_params,
                                             aligned_data,
                                             force_continuous=force_continuous,
                                             static_features=static_features,
                                             verbose=verbose,
                                             bad_frames=bad_frames,
                                             reconstruct_spikes=reconstruct_spikes)

            save_exp_to_pickle(Exp, exppath, verbose=verbose)
            return Exp, load_log

        else:
            raise ValueError('External data sources not supported yet')


def save_exp_to_pickle(exp, path, verbose=True):
    with open(path, "wb") as f:
        pickle.dump(exp, f)
        if verbose:
            print(f'Experiment {exp.signature} saved to {path}\n')


def load_exp_from_pickle(path, verbose=True):
    with open(path, "rb") as f:
        exp = pickle.load(f,)
        if verbose:
            print(f'Experiment {exp.signature} loaded from {path}\n')
    return exp



from .matrix_utils import *
import scipy
from scipy.linalg import expm
import math

def renyi_divergence(A, B, q):
    if q <= 0:
        raise Exception('q must be >0')
    elif q == 1:
        answer = np.trace(np.dot(A, (scipy.linalg.logm(A) - scipy.linalg.logm(B))/np.log(2.0)))
    else:
        answer = (1/(q-1)) * np.log(np.trace(np.dot(scipy.linalg.fractional_matrix_power(A, q),
                                                    scipy.linalg.fractional_matrix_power(B, 1-q))))/np.log(2.0)
    return answer


def get_density_matrix(A, t, norm=0):
    A = A.astype(float)
    if norm:
        X = get_norm_laplacian(A)
    else:
        X = get_laplacian(A)

    R = expm(-t * X)
    R = R/np.trace(X)

    return R


def manual_entropy(pr):
    probs = np.trim_zeros(pr)
    probs = probs[np.where(probs > 1e-15)]
    return -np.real(np.sum(np.multiply(probs, np.log2(probs))))


def js_divergence(A, B, t, return_partial_entropies=True):
    X = get_density_matrix(A, t)
    Y = get_density_matrix(B, t)

    mixed = np.trim_zeros(np.linalg.eigvalsh((X + Y) / 2))
    raw1 = np.trim_zeros(np.linalg.eigvalsh(X))
    raw2 = np.trim_zeros(np.linalg.eigvalsh(Y))

    first = manual_entropy(mixed)
    ent1, ent2 = manual_entropy(raw1), manual_entropy(raw2)
    second = 0.5*(ent1 + ent2)
    try:
        JSD = math.sqrt(first - second)
    except:
        JSD = 0

    if not return_partial_entropies:
        return JSD
    else:
        return manual_entropy(mixed), ent1, ent2, JSD


import numpy as np
from src.driada.experiment.neuron import *

def test_init():
    pseudo_ca = np.random.random(size=10000)
    pseudo_sp = np.zeros(10000)
    neuron = Neuron(0, pseudo_ca, pseudo_sp)



from .utils import *
from scipy.stats import entropy
from ..utils.data import correlation_matrix

DATA_SHAPE_THR = 0.01  # if n/t in multivariate time series data is more than DATA_SHAPE_THR,
                      # cov/corr spectrum may be significantly distorted, correction is recommended


def _eff_dim(corr_eigs, q=2):
    if q < 0:
        raise ValueError('Renyi entropy is undefined for q<0')

    norm_corr_eigs = corr_eigs / np.sum(corr_eigs)
    if q == 1:
        # standard entropy
        return 2**entropy(norm_corr_eigs, base=2)

    elif q == 2:
        # quadratic entropy
        return (sum(corr_eigs)**2)/sum([e**2 for e in corr_eigs])

    elif q == np.inf:
        # min-entropy
        return np.sum(corr_eigs)/np.max(corr_eigs)

    else:
        return 1.0/(1.0 - q)*np.log(np.sum([p**q for p in norm_corr_eigs]))


def eff_dim(data, enable_correction, q=2, **correction_kwargs):
    n, t = data.shape
    if 1.0*n/t > DATA_SHAPE_THR and not enable_correction:
        raise UserWarning(f'fN/T is {1.0*n/t}, which is bigger than {DATA_SHAPE_THR}. Spectrum correction is recommended')

    cmat = correlation_matrix(data)
    if enable_correction:
        corrected_eigs = correct_cov_spectrum(n, t, cmat, **correction_kwargs)
        final_eigs = corrected_eigs[-1]
    else:
        final_eigs = eigh(cmat, eigvals_only=True)

    return _eff_dim(final_eigs, q=q)






from .graph_utils import *
from .matrix_utils import *
import warnings
import random

def adj_random_rewiring_iom_preserving(a, is_weighted, r=10):
    # print ('Rewiring double connections...')

    s = symmetric_component(a, is_weighted)
    rs = turn_to_partially_directed(s, directed=1.0, weighted=is_weighted)
    rows, cols = rs.A.nonzero()
    edgeset = set(zip(rows, cols))
    upper = [l for l in edgeset]  # if l[0]<l[1]]
    source_nodes = [e[0] for e in upper]
    target_nodes = [e[1] for e in upper]

    double_edges = len(upper)

    i = 0

    while i < double_edges * r:
        # print('i=',i)
        good_choice = 0
        while not good_choice:
            ind1, ind2 = np.random.choice(double_edges, 2)
            n1, n3 = source_nodes[ind1], source_nodes[ind2]
            n2, n4 = target_nodes[ind1], target_nodes[ind2]

            if len(set([n1, n2, n3, n4])) == 4:
                good_choice = 1

        w1 = s[n1, n2]
        w2 = s[n2, n1]
        w3 = s[n3, n4]
        w4 = s[n4, n3]

        '''
        if w1*w2*w3*w4 == 0:
            print(i, w1, w2, w3, w4)
        '''
        if s[n1, n3] + s[n1, n4] + s[n2, n3] + s[n2, n4] == 0:
            s[n1, n4] = w1
            s[n4, n1] = w2
            s[n2, n3] = w3
            s[n3, n2] = w4

            s[n1, n2] = 0
            s[n2, n1] = 0
            s[n3, n4] = 0
            s[n4, n3] = 0

            target_nodes[ind1], target_nodes[ind2] = n4, n2

            i += 1
            # print(get_symmetry_index(sp.csr_array(A)))

    # plt.matshow(s)
    # print ('Rewiring single connections...')

    ns = non_symmetric_component(a, is_weighted)

    # plt.matshow(ns)
    rows, cols = ns.nonzero()
    edges = list((set(zip(rows, cols))))
    source_nodes = [e[0] for e in edges]
    target_nodes = [e[1] for e in edges]
    single_edges = len(edges)

    i = 0

    while i < single_edges * r:
        # while i < 10:
        good_choice = 0
        while not good_choice:
            ind1, ind2 = np.random.choice(single_edges, 2)
            n1, n3 = source_nodes[ind1], source_nodes[ind2]
            n2, n4 = target_nodes[ind1], target_nodes[ind2]

            if len(set([n1, n2, n3, n4])) == 4:
                good_choice = 1

        w1 = ns[n1, n2]
        w2 = ns[n3, n4]

        checklist = [ns[n1, n3], ns[n1, n4], ns[n2, n3], ns[n2, n4],
                     ns[n3, n1], ns[n4, n1], ns[n3, n2], ns[n4, n2],
                     s[n3, n1], s[n4, n1], s[n3, n2], s[n4, n2]]

        if checklist.count(0) == 12:
            ns[n1, n4] = w1
            ns[n3, n2] = w2

            ns[n1, n2] = 0
            ns[n3, n4] = 0

            i += 1

            target_nodes[ind1], target_nodes[ind2] = n4, n2
            # print(get_symmetry_index(sp.csr_array(A)))

    res = s + ns
    if not is_weighted:
        res = res.astype(bool)

    return sp.csr_array(res)


def random_rewiring_complete_graph(a, p=1.0):
    # p ranges from 0 to 1 and defines the degree of reshuffling (percent of edges affected)
    n = a.shape[0]
    if len(np.nonzero(a)[0]) < n ** 2 - n:
        raise Exception('Graph is not complete')

    symmetric = np.allclose(a, a.T)

    all_x_inds, all_y_inds = np.nonzero(a)
    vals = a[all_x_inds, all_y_inds]
    shuffled_positions = np.random.choice(np.arange((n ** 2 - n)),
                                          replace=False,
                                          size=int(p * (n ** 2 - n)))

    shuffled_x_inds = all_x_inds[shuffled_positions]
    shuffled_y_inds = all_y_inds[shuffled_positions]
    shuffled_vals = vals[shuffled_positions]
    np.random.shuffle(shuffled_vals)  # shuffling in-place

    stable_part = a.copy()
    if symmetric:
        stable_part[shuffled_x_inds, shuffled_y_inds] = 0
        stable_part[shuffled_y_inds, shuffled_x_inds] = 0
    else:
        stable_part[shuffled_x_inds, shuffled_y_inds] = 0

    if symmetric:
        shuffled_half = np.zeros(a.shape)
        shuffled_half[shuffled_x_inds, shuffled_y_inds] = shuffled_vals
        shuffled_part = (shuffled_half + shuffled_half.T) / 2.0
    else:
        shuffled_part = np.zeros(a.shape)
        shuffled_part[shuffled_x_inds, shuffled_y_inds] = shuffled_vals

    rewired = stable_part + shuffled_part

    return rewired


def random_rewiring_dense_graph(a):
    if isinstance(a, np.ndarray):
        afull = a
        nelem = len(np.nonzero(a)[0])
    else:
        afull = a.A
        nelem = a.nnz

    if nelem != a.shape[0] ** 2 - a.shape[0]:
        # warnings.warn('Graph is not complete, proceeding with gap filling trick')
        temp = 0.0001
    else:
        temp = 0

    psA = afull + np.full(a.shape, temp) - np.eye(a.shape[0]) * temp
    vals = psA[np.nonzero(np.triu(psA))]
    np.random.shuffle(vals)

    tri = np.zeros(a.shape)
    tri[np.triu_indices(a.shape[0], 1)] = vals
    rewired = tri + tri.T
    res = rewired - np.full(a.shape, temp) + np.eye(a.shape[0]) * temp

    return res


def get_single_double_edges_lists(g):
    L1 = []
    L2 = []
    h = nx.to_undirected(g).copy()
    for e in h.edges():
        if g.has_edge(e[1], e[0]):
            if g.has_edge(e[0], e[1]):
                L2.append((e[0], e[1]))
            else:
                L1.append((e[1], e[0]))
        else:
            L1.append((e[0], e[1]))

    return [L1, L2]


# warning: the code below is legacy and may be inefficient
def random_rewiring_IOM_preserving(G, r=10):

    [L1, L2] = get_single_double_edges_lists(G)
    Number_of_single_edges = len(L1)
    Number_of_double_edges = len(L2)
    Number_of_rewired_1_edge_pairs = Number_of_single_edges * r
    Number_of_rewired_2_edge_pairs = Number_of_double_edges * r
    # Number_of_rewired_2_edge_pairs = 20
    i = 0
    count = 0
    previous_text = ''

    # print len(set(L).intersection(List_of_edges))
    print('Rewiring double connections...')
    while i < Number_of_rewired_2_edge_pairs:
        Edge_index_1 = random.randint(0, Number_of_double_edges - 1)
        Edge_index_2 = random.randint(0, Number_of_double_edges - 1)
        Edge_1 = L2[Edge_index_1]
        Edge_2 = L2[Edge_index_2]
        [Node_A, Node_B] = Edge_1
        [Node_C, Node_D] = Edge_2
        while (Node_A == Node_C) or (Node_A == Node_D) or (Node_B == Node_C) or (Node_B == Node_D):
            Edge_index_1 = random.randint(0, Number_of_double_edges - 1)
            Edge_index_2 = random.randint(0, Number_of_double_edges - 1)
            Edge_1 = L2[Edge_index_1]
            Edge_2 = L2[Edge_index_2]
            [Node_A, Node_B] = Edge_1
            [Node_C, Node_D] = Edge_2

        # print ('Edges:',Node_A, Node_B, ';',Node_C, Node_D)
        # print G.has_edge(Node_A, Node_B), G.has_edge(Node_B, Node_A), G.has_edge(Node_C, Node_D), G.has_edge(Node_D, Node_C)
        if G.has_edge(Node_A, Node_D) == 0 and G.has_edge(Node_D, Node_A) == 0 and G.has_edge(Node_C,
                                                                                              Node_B) == 0 and G.has_edge(
                Node_B, Node_C) == 0:
            # try:
            try:
                w_ab = G.get_edge_data(Node_A, Node_B)['weight']
            except:
                pass
            G.remove_edge(Node_A, Node_B)
            G.remove_edge(Node_B, Node_A)
            '''
            except nx.NetworkXError:
                pass
                #print('fuck')
            '''
            try:
                try:
                    w_cd = G.get_edge_data(Node_C, Node_D)['weight']
                except:
                    pass
                G.remove_edge(Node_C, Node_D)
                G.remove_edge(Node_D, Node_C)
            except nx.NetworkXError:
                pass
                # print('fuck')

            try:
                G.add_edge(Node_A, Node_D, weight=w_ab)
                G.add_edge(Node_D, Node_A, weight=w_ab)
            except:
                G.add_edge(Node_A, Node_D)
                G.add_edge(Node_D, Node_A)

            try:
                G.add_edge(Node_C, Node_B, weight=w_cd)
                G.add_edge(Node_B, Node_C, weight=w_cd)
            except:
                G.add_edge(Node_C, Node_B)
                G.add_edge(Node_B, Node_C)

            # print L2[Edge_index_1]
            L2[Edge_index_1] = (Node_A, Node_D)
            # print L2[Edge_index_1]
            # L2[Edge_index_1+1] = (Node_D, Node_A)
            L2[Edge_index_2] = (Node_C, Node_B)
            # L2[Edge_index_2+1] = (Node_B, Node_C)
            i += 1

        if (i != 0) and (i % (Number_of_double_edges // 1)) == 0:
            text = str(round(100.0 * i / Number_of_rewired_2_edge_pairs, 0)) + "%"
            if text != previous_text:
                # print text
                pass
            previous_text = text

    i = 0
    print('Rewiring single connections...')
    while i < Number_of_rewired_1_edge_pairs:
        Edge_index_1 = random.randint(0, Number_of_single_edges - 1)
        Edge_index_2 = random.randint(0, Number_of_single_edges - 1)
        Edge_1 = L1[Edge_index_1]
        Edge_2 = L1[Edge_index_2]
        [Node_A, Node_B] = Edge_1
        [Node_C, Node_D] = Edge_2
        while (Node_A == Node_C) or (Node_A == Node_D) or (Node_B == Node_C) or (Node_B == Node_D):
            Edge_index_1 = random.randint(0, Number_of_single_edges - 1)
            Edge_index_2 = random.randint(0, Number_of_single_edges - 1)
            Edge_1 = L1[Edge_index_1]
            Edge_2 = L1[Edge_index_2]
            [Node_A, Node_B] = Edge_1
            [Node_C, Node_D] = Edge_2

        if G.has_edge(Node_A, Node_D) == 0 and G.has_edge(Node_D, Node_A) == 0 and G.has_edge(Node_C,
                                                                                              Node_B) == 0 and G.has_edge(
                Node_B, Node_C) == 0:
            try:
                try:
                    w_ab = G.get_edge_data(Node_A, Node_B)['weight']
                except:
                    pass
                G.remove_edge(Node_A, Node_B)


            except nx.NetworkXError:
                print('fuck')

            try:
                try:
                    w_cd = G.get_edge_data(Node_C, Node_D)['weight']
                except:
                    pass
                G.remove_edge(Node_C, Node_D)

            except nx.NetworkXError:
                print('fuck')

            try:
                G.add_edge(Node_A, Node_D, weight=w_ab)
            except:
                G.add_edge(Node_A, Node_D)

            try:
                G.add_edge(Node_C, Node_B, weight=w_cd)
            except:
                G.add_edge(Node_C, Node_B)

            L1[Edge_index_1] = (Node_A, Node_D)
            L1[Edge_index_2] = (Node_C, Node_B)
            i += 1

        if (i != 0) and (i % (Number_of_single_edges // 1)) == 0:
            text = str(round(100.0 * i / Number_of_rewired_1_edge_pairs, 0)) + "%"
            if text != previous_text:
                # print text
                pass
            previous_text = text

    G_rewired = copy.deepcopy(G)

    return G_rewired


import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
# import torchvision
from torch.utils.data import Dataset, DataLoader


class Encoder(nn.Module):

    def __init__(self, orig_dim, inter_dim, code_dim, kwargs, device=None):
        super().__init__()
        dropout = kwargs.get('dropout', None)

        self.encoder_hidden_layer = nn.Linear(
            in_features=orig_dim, out_features=inter_dim
        )
        self.encoder_output_layer = nn.Linear(
            in_features=inter_dim, out_features=code_dim
        )

        if dropout is not None:
            if 0 <= dropout < 1:
                self.dropout = nn.Dropout(p=dropout)
            else:
                raise ValueError('Dropout rate should be in the range 0<=dropout<1')
        else:
            self.dropout = nn.Dropout(0.0)

        if device is None:
            self._device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        else:
            self._device = device

    def forward(self, features):
        activation = self.encoder_hidden_layer(features)
        activation = self.dropout(torch.ones(activation.shape).to(self._device)) * activation
        activation = F.leaky_relu(activation)
        # activation = torch.relu(activation)
        code = self.encoder_output_layer(activation)
        code = torch.sigmoid(code)
        # code = F.leaky_relu(code)

        return code


class Decoder(nn.Module):

    def __init__(self, code_dim, inter_dim, orig_dim, kwargs, device=None):
        super().__init__()
        dropout = kwargs.get('dropout', None)

        self.decoder_hidden_layer = nn.Linear(
            in_features=code_dim, out_features=inter_dim
        )
        self.decoder_output_layer = nn.Linear(
            in_features=inter_dim, out_features=orig_dim
        )

        if dropout is not None:
            if 0 <= dropout < 1:
                self.dropout = nn.Dropout(p=dropout)
            else:
                raise ValueError('Dropout rate should be in the range 0<=dropout<1')
        else:
            self.dropout = nn.Dropout(0.0)

        if device is None:
            self._device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        else:
            self._device = device

    def forward(self, code):
        activation = self.decoder_hidden_layer(code)
        # activation = torch.relu(activation)
        activation = self.dropout(torch.ones(activation.shape).to(self._device)) * activation
        activation = F.leaky_relu(activation)
        activation = self.decoder_output_layer(activation)
        reconstructed = torch.sigmoid(activation)
        # reconstructed = F.leaky_relu(activation)

        return reconstructed


class AE(nn.Module):

    def __init__(self, orig_dim, inter_dim, code_dim, enc_kwargs=None, dec_kwargs=None, device=None):
        super().__init__()
        if device is None:
            self._device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        else:
            self._device = device
        self.encoder = Encoder(orig_dim=orig_dim, inter_dim=inter_dim, code_dim=code_dim,
                               kwargs=enc_kwargs, device=self._device)
        self.decoder = Decoder(orig_dim=orig_dim, inter_dim=inter_dim, code_dim=code_dim,
                               kwargs=dec_kwargs, device=self._device)

    def forward(self, features):
        code = self.encoder.forward(features)
        reconstructed = self.decoder.forward(code)
        return reconstructed

    def get_code_embedding(self, input_):
        encoder = self.encoder
        embedding = encoder.forward(input_)
        return embedding.detach().cpu().numpy().T


class VAE(nn.Module):

    def __init__(self, orig_dim, inter_dim, code_dim, enc_kwargs=None, dec_kwargs=None):
        super(VAE, self).__init__()

        self.encoder = Encoder(orig_dim=orig_dim, inter_dim=inter_dim, code_dim=2 * code_dim, kwargs=enc_kwargs)
        self.decoder = Decoder(orig_dim=orig_dim, inter_dim=inter_dim, code_dim=code_dim, kwargs=dec_kwargs)
        self.orig_dim = orig_dim
        self.inter_dim = inter_dim
        self.code_dim = code_dim

    def reparameterization(self, mu, log_var):
        """
        :param mu: mean from the encoder's latent space
        :param log_var: log variance from the encoder's latent space
        """
        std = torch.exp(0.5 * log_var)  # standard deviation
        eps = torch.randn_like(std)  # `randn_like` as we need the same size
        sample = mu + (eps * std)  # sampling as if coming from the input space
        return sample

    def get_code(self, features):
        x = self.encoder.forward(features)

        # print('x shape:', x.shape)
        x = x.view(-1, 2, self.code_dim)

        # get `mu` and `log_var`
        mu = x[:, 0, :]  # the first feature values as mean
        log_var = x[:, 1, :]  # the other feature values as variance

        # print('mu shape:', mu.shape)
        # get the latent vector through reparameterization
        code = self.reparameterization(mu, log_var)
        # print('code shape:', mu.shape)

        return code, mu, log_var

    def forward(self, features):
        # encoding
        code, mu, log_var = self.get_code(features)

        # decoding
        reconstructed = self.decoder.forward(code)
        return reconstructed, mu, log_var

    def get_code_embedding(self, input_):
        #encoder = self.encoder
        embedding, mu, log_var = self.get_code(input_)
        return embedding.detach().cpu().numpy().T


class NeuroDataset(Dataset):
    """Neural activity dataset."""

    def __init__(self, data, transform=None):

        self.data = data.T
        self.transform = transform

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()

        sample = {'vector': self.data[idx].reshape(-1, 1), 'target': 0}

        if self.transform:
            sample = self.transform(sample)

        return self.data[idx], -42, idx
        # return sample



from scipy.spatial.distance import pdist, cdist
from scipy.sparse.csgraph import shortest_path
from scipy.stats import pearsonr, norm
from scipy.linalg import eigh
import numpy as np


def res_var_metric(all_dists, emb_dists):
    m = 1 - (pearsonr(all_dists, emb_dists)[0])**2
    return m


def correct_cov_spectrum(N, T, cmat, correction_iters=10, ensemble_size=1):
    eigs = eigh(cmat, eigvals_only=True)

    init_eigs = eigs.copy()
    iter_eigs = eigs.copy()
    corrected_eigs = [init_eigs]

    for i in range(correction_iters):
        all_ratios = np.zeros((ensemble_size, N))

        # phase 1
        for j in range(ensemble_size):
            M = norm.rvs(size=(N,T))
            L = np.diag(np.sqrt(iter_eigs))
            M2 = L@M@M.T@L/T
            ps_eigs = eigh(M2)[0]
            all_ratios[j,:] = np.divide(ps_eigs, iter_eigs)

        s1 = np.sum(all_ratios, axis=0)
        s2 = np.sum(np.square(all_ratios), axis=0)
        S = np.diag(np.divide(s1, s2))

        iter_eigs = eigh(np.diag(init_eigs)@S, eigvals_only=True)

        # phase 2
        M = norm.rvs(size=(N,T))
        L = np.diag(np.sqrt(iter_eigs))
        W = L@M@M.T@L/T
        _, V = eigh(W)
        upd_eigs = np.diagonal(V@np.diag(init_eigs)@V.T)

        corrected_eigs.append(upd_eigs)
        iter_eigs = upd_eigs

    return corrected_eigs



import numpy as np
import matplotlib.pyplot as plt
import scipy.sparse as sp

from .dr_base import *
from ..utils.data import correlation_matrix, to_numpy_array, rescale
from .embedding import Embedding
from .graph import ProximityGraph

# TODO: refactor this
def check_data_for_errors(d):
    sums = np.sum(np.abs(d), axis=0)
    if len(sums.nonzero()[1]) != d.shape[1]:
        bad_points = np.where(sums == 0)[1]
        print('zero points:', bad_points)
        print(d.A[:, bad_points[0]])
        raise Exception('Data contains zero points!')


class MVData(object):
    '''
    Main class for multivariate data storage & processing
    '''

    def __init__(self,
                 data,
                 labels=None,
                 distmat=None,
                 rescale_rows=False,
                 data_name=None,
                 downsampling=None):

        if downsampling is None:
            self.ds = 1
        else:
            self.ds = int(downsampling)

        self.data = to_numpy_array(data)[:, ::self.ds]

        # TODO: add support for various preprocessing methods (wvt, med_filt, etc.)
        self.rescale_rows = rescale_rows
        if self.rescale_rows:
            for i, row in enumerate(self.data):
                self.data[i] = rescale(row)

        self.data_name = data_name
        self.n_dim = self.data.shape[0]
        self.n_points = self.data.shape[1]

        if labels is None:
            self.labels = np.zeros(self.n_points)
        else:
            self.labels = to_numpy_array(labels)

        self.distmat = distmat

    def median_filter(self, window):
        from scipy.signal import medfilt
        d = self.data.A

        new_d = medfilt(d, window)

        self.data = sp.csr_matrix(new_d)

    def corr_mat(self):
        cm = correlation_matrix(self.data)
        return cm

    def get_distmat(self, method='euclidean'):
        raise NotImplementedError('To be implemented soon')

    def get_embedding(self, e_params, g_params=None, m_params=None, kwargs=None):
        method = e_params['e_method']
        method_name = e_params['e_method_name']

        if method_name not in EMBEDDING_CONSTRUCTION_METHODS:
            raise Exception('Unknown embedding construction method!')

        graph = None
        if method.requires_graph:
            if g_params is None:
                raise ValueError(f'Method {method_name} requires proximity graph, but '
                                 f'graph params were not provided')
            if g_params['weighted'] and m_params is None:
                raise ValueError(f'Method {method_name} requires weights for proximity graph, but '
                                 f'metric params were not provided')

            graph = self.get_proximity_graph(m_params, g_params)

        if method.requires_distmat and self.distmat is None:
            raise Exception(f'No distmat provided for {method_name} method.'
                            f' Try constructing it first with get_distmat() method')

        emb = Embedding(self.data, self.distmat, self.labels, e_params, g=graph)
        emb.build(kwargs=kwargs)

        return emb

    def get_proximity_graph(self, m_params, g_params):
        if g_params['g_method_name'] not in GRAPH_CONSTRUCTION_METHODS:
            raise Exception('Unknown graph construction method!')

        graph = ProximityGraph(self.data, m_params, g_params)
        # print('Graph succesfully constructed')
        return graph

    def draw_vector(self, num):
        data = self.data[:, num]
        plt.matshow(data.reshape(1, self.n_dim))
        plt.matshow(self.data)

    def draw_row(self, num):
        data = self.data[num, :]
        plt.figure(figsize=(12, 10))
        plt.plot(data)



import shutil
import importlib
import git
import sys
import os
from os.path import join

def clone_org_repo(repo, org='iabs-neuro', branch='main', path_to_clone=None):
    rp = git.Repo.clone_from(f'https://github.com/{org}/{repo}',
                             branch=branch,
                             to_path=path_to_clone)


def reload_module(name, exmodule, absname=None):
    print(f'module "{name}" is already imported, reloading...')
    exmodpath = exmodule.__path__[0]
    shutil.rmtree(exmodpath)
    print(f'Cloning module "{name}" to {exmodpath}...')
    clone_org_repo(name.split('.')[-1], path_to_clone=exmodpath)
    importlib.reload(exmodule)
    print(f'module "{name}" is successfully reloaded')


def import_external_repositories(repolist, extpath='/content', reload_internal=False):
    os.makedirs(extpath, exist_ok=True)

    for repo in repolist:
        rppath = os.path.join(extpath, repo)

        if repo in sys.modules:
            exmodule = sys.modules[repo]
            reload_module(repo, exmodule)

            if reload_internal:
                for module_name in os.listdir(f'{repo}/externals'):
                    intmodname = f'{repo}.externals.{module_name}'
                    intmodule = sys.modules[intmodname]
                    reload_module(intmodname, intmodule)

        else:
            if repo in os.listdir(extpath):
                shutil.rmtree(join(extpath, repo))
            print(f'Cloning module "{repo}" to {extpath}...')
            clone_org_repo(repo, path_to_clone=rppath)
            importlib.import_module(repo, package=None)


# @title Graph class { form-width: "200px" }

import pynndescent
import scipy.sparse as sp
import numpy as np
import matplotlib.pyplot as plt
import numpy.random as npr

from scipy.optimize import curve_fit

from sklearn.neighbors import kneighbors_graph
from sklearn.utils.validation import check_symmetric
from scipy.sparse.csgraph import shortest_path
from umap.umap_ import fuzzy_simplicial_set
from .dr_base import *

from src.driada.network.net_base import Network


class ProximityGraph(Network):
    '''
    Graph built on data points which represents the underlying manifold
    '''

    def __init__(self, d, m_params, g_params, create_nx_graph=False):
        self.all_metric_params = m_param_filter(m_params)
        self.metric = m_params['metric_name']
        self.metric_args = {key: self.all_metric_params[key] for key in self.all_metric_params.keys()
                            if key not in ['metric_name', 'sigma']}

        all_params = g_param_filter(g_params)
        for key in all_params:
            setattr(self, key, g_params[key])

        self.data = d

        self.construct_adjacency()
        # TODO: add graph_preprocessing to changeable graph params
        super(ProximityGraph, self).__init__(adj=self.adj,
                                             preprocessing='giant_cc',
                                             create_nx_graph=create_nx_graph,
                                             directed=False,
                                             weighted=all_params['weighted'])

        node_mapping = self._init_to_final_node_mapping
        lost_nodes = set(range(self.n)) - set(list(node_mapping.keys()))
        if len(lost_nodes) > 0:
            print(f'{len(lost_nodes)} nodes lost after giant component creation!')

            if len(lost_nodes) >= self.max_deleted_nodes * self.n:
                raise Exception(f'more than {self.max_deleted_nodes * 100} % of nodes discarded during gc creation!')
            else:
                self.lost_nodes = lost_nodes
                connected = [i for i in range(self.n) if i not in self.lost_nodes]
                self.bin_adj = self.bin_adj[connected, :].tocsc()[:, connected].tocsr()
                self.neigh_distmat = self.neigh_distmat[connected, :].tocsc()[:, connected].tocsr()

        self._checkpoint()

        #arr = np.array(range(self.n)).reshape(-1, 1)
        #self.timediff = cdist(arr, arr, 'cityblock')
        #self.norm_timediff = self.timediff / (self.n / 3)

    def distances_to_affinities(self):
        if self.neigh_distmat is None:
            raise Exception('distances between nearest neighbors not available')

        if not self.weighted:
            raise Exception('no need to construct affinities for binary graph weights')

        if self.dist_to_aff == 'hk':
            sigma = self.all_metric_params['sigma']
            self.adj = self.neigh_distmat.copy()
            sqdist_matrix = self.neigh_distmat.multiply(self.neigh_distmat)
            mean_sqdist = sqdist_matrix.sum() / sqdist_matrix.nnz
            self.adj.data = np.exp(-sqdist_matrix.data / (1.0 * sigma * mean_sqdist))

    def construct_adjacency(self):
        construct_fn = getattr(self, 'create_' + self.g_method_name + '_graph_')
        construct_fn()

    def _checkpoint(self):
        if self.adj is not None:
            if not sp.issparse(self.adj):
                # check for sparsity violation
                raise Exception('Adjacency matrix is not sparse!')
            self.adj = check_symmetric(self.adj, raise_exception=True)
            self.bin_adj = check_symmetric(self.bin_adj, raise_exception=True)
            self.neigh_distmat = check_symmetric(self.neigh_distmat, raise_exception=True)
            print('Adjacency symmetry confirmed')

        else:
            raise Exception('Adjacency matrix is not constructed, checkpoint routines are unavailable')

    def create_umap_graph_(self):
        RAND = np.random.RandomState(42)
        adj, _, _, dists = fuzzy_simplicial_set(self.data.T, self.nn, metric=self.metric,
                                                metric_kwds=self.metric_args,
                                                random_state=RAND,
                                                return_dists=True)

        self.adj = sp.csr_matrix(adj)
        #print('WARNING: distmat is not yet implemented in this branch')

    def create_knn_graph_(self):
        if self.metric in named_distances:
            curr_metric = self.metric
        else:
            curr_metric = globals()[self.metric]

        N = self.data.shape[1]
        index = pynndescent.NNDescent(self.data.T,
                                      metric=curr_metric,
                                      metric_kwds=self.metric_args,
                                      n_neighbors=self.nn + 1,
                                      diversify_prob=1.0,
                                      pruning_degree_multiplier=1.5)

        neighs, dists = index.neighbor_graph

        neigh_cols = neighs[:, 1:].flatten()
        dist_vals = dists[:, 1:].flatten()
        neigh_rows = np.repeat(np.arange(N), self.nn)

        all_neigh_cols = np.concatenate((neigh_cols, neigh_rows))
        all_neigh_rows = np.concatenate((neigh_rows, neigh_cols))
        all_dist_vals = np.concatenate((dist_vals, dist_vals))

        self.bin_adj = sp.csr_matrix((np.array([True for _ in range(2 * N * self.nn)]),
                                      (all_neigh_rows, all_neigh_cols)))

        self.neigh_distmat = sp.csr_matrix((all_dist_vals, (all_neigh_rows, all_neigh_cols)))

        if self.weighted:
            self.distances_to_affinities()
        else:
            self.adj = self.bin_adj.copy()

    def create_auto_knn_graph_(self):
        A = kneighbors_graph(self.data.T, self.nn, mode='connectivity', include_self=False)
        A.setdiag(0)
        A = A.astype(bool, casting='unsafe', copy=True)
        A = A + A.T
        self.adj = A
        #print('WARNING: distmat is not yet implemented in this branch')

    def calculate_indim(self, mode, factor=2):

        def normalizing_const(dim):
            if dim == 0:
                return 1
            if dim == 1:
                return 2
            elif dim == 2:
                return np.pi / 2
            else:
                return (dim - 1.0) / dim * normalizing_const(dim - 2)

        def func(x, a):
            # C = normalizing_const(a)
            return a * (np.log10(np.sin(1.0 * x * np.pi / 2)))  # + np.log10(C)
            # return a*(np.log10(np.sin(1.0*x/xmax*np.pi/2))) - np.log10(valmax)

        def func2(x, a):
            return -a / 2. * (x - max(x)) ** 2

        if self.g_method not in ['w_knn', 'w_auto_knn']:
            raise Exception('Distance matrix construction missed!')

        print('Calculating graph internal dimension...')

        if mode == 'fast':
            distmat = sp.csr_matrix(self.sqdist_matrix)
            distmat.data = np.sqrt(self.sqdist_matrix.data)
            indices = list(npr.permutation(npr.choice(self.n, size=self.n // factor, replace=False)))
            dm = distmat[indices, :][:, indices]

        elif mode == 'full':
            dm = self.sqdist_matrix
            dm.data = np.sqrt(self.sqdist_matrix.data)

        print('Shortest path computation started, distance matrix size: ', dm.shape)
        spmatrix = shortest_path(dm, method='D', directed=False)
        all_dists = spmatrix.flatten()
        all_dists = all_dists[all_dists != 0]

        nbins = 500
        pre_hist = np.histogram(all_dists, bins=nbins, density=True)
        # plt.hist(all_dists, bins = 500)
        dx = pre_hist[1][1] - pre_hist[1][0]
        dmax_bin = np.argmax(pre_hist[0])
        dmax = pre_hist[1][dmax_bin] + dx / 2.0

        hist = np.histogram(all_dists / dmax, bins=nbins, density=True)
        distr_x = hist[1][:-1] + dx / 2
        distr_y = hist[0] / max(hist[0][0:nbins])
        avg = np.mean(all_dists)
        std = np.std(all_dists / dmax)

        res = []
        # print(distr_x)
        # print(distr_y)
        left_distr_x = distr_x[(distr_x > 1 - 2. * std) & (distr_x <= 1) & (distr_y > 1e-6)]
        left_distr_y = np.log10(distr_y[(distr_x[:] > 1 - 2. * std) & (distr_x[:] <= 1)])  # & (distr_y[:]>1e-6)])

        for D in [0.1 * x for x in range(10, 260)]:
            y = func(left_distr_x, D - 1)
            res.append(np.linalg.norm(y - left_distr_y) / np.sqrt(len(y)))

        plot = 0
        if plot:
            fig = plt.figure(2, figsize=(12, 10))
            plt.plot(np.linspace(0, len(res) / 10.0, num=len(res)), res)

        Dmin = 0.1 * (np.argmax(-np.array(res)) + 1)
        print('Dmin = ', Dmin)
        fit = curve_fit(func2, left_distr_x, left_distr_y)
        # print(fit)
        a = (fit[0][0])
        # print('Dfit = ', Dfit)

        plot = 0
        if plot:
            fig = plt.figure(1, figsize=(12, 10))
            ax = fig.add_subplot(111)
            plt.hist(all_dists / dmax, bins=nbins, histtype='stepfilled', density=True, log=True)

        alpha = 2.0
        R = np.sqrt(2 * a)
        print('R = ', R)
        Dpr = 1 - alpha ** 2 / (2 * np.log(np.cos(alpha * np.pi / 2.0 / R)))
        print('D_calc = ', Dpr)

        return Dmin, Dpr

    def scaling(self):
        mat = self.adj.astype(bool).A.astype(int)
        diagsums = []
        for i in range(self.n - 1):
            diagsums.append(np.trace(mat, offset=i, dtype=None, out=None) / (self.n - i))

        return diagsums




from scipy.sparse.linalg import eigs
import umap.umap_ as umap
from pydiffmap import diffusion_map as dm
from scipy.sparse.csgraph import shortest_path

from sklearn.decomposition import PCA
from sklearn.manifold import spectral_embedding, Isomap, LocallyLinearEmbedding, TSNE
# from sklearn.cluster.spectral import discretize

# warnings.filterwarnings("ignore")

from .dr_base import *
from .graph import ProximityGraph
from .neural import *
from .mvu import *

from ..network.matrix_utils import get_inv_sqrt_diag_matrix

def norm_cross_corr(a, b):
    a = (a - np.mean(a)) / (np.std(a) * len(a))
    b = (b - np.mean(b)) / (np.std(b))
    c = np.correlate(a, b, 'full')
    return c


def remove_outliers(data, thr_percentile):
    thr1 = np.mean(data) + thr_percentile * np.std(data)
    thr2 = np.mean(data) - thr_percentile * np.std(data)
    good_points = np.where((data < thr1) & (data > thr2))[0]

    return good_points, data[good_points]


class Embedding:
    '''
    Low-dimensional representation of data
    '''

    def __init__(self, init_data, init_distmat, labels, params, g=None):
        if g is not None:
            if isinstance(g, ProximityGraph):
                self.graph = g
            else:
                raise Exception('Wrong graph type!')

        self.all_params = e_param_filter(params)
        for key in params:
            setattr(self, key, params[key])

        self.init_data = init_data
        self.init_distmat = init_distmat

        if self.e_method.is_linear:
            self.transformation_matrix = None

        self.labels = labels
        self.coords = None

        try:
            self.nclasses = len(set(self.labels))
        except:
            self.nclasses = np.unique(self.labels)

        if self.e_method.nn_based:
            self.nnmodel = None

    def build(self, kwargs=None):
        if kwargs is None:
            kwargs = dict()
        fn = getattr(self, 'create_' + self.e_method_name + '_embedding_')

        if self.e_method.requires_graph:
            # TODO: move connectivity check to graph
            if not self.graph.is_connected():
                raise Exception('Graph is not connected!')

        fn(**kwargs)

    def create_pca_embedding_(self, verbose=True):
        if verbose:
            print('Calculating PCA embedding...')

        pca = PCA(n_components=self.dim)
        self.coords = pca.fit_transform(self.init_data.T).T
        self.reducer_ = pca
        # print(pca.explained_variance_ratio_)

    def create_isomap_embedding_(self):
        A = self.graph.adj
        map = Isomap(n_components=self.dim, n_neighbors=self.graph.nn,
                     metric='precomputed')
        # self.coords = sp.csr_matrix(map.fit_transform(self.graph.data.A.T).T)
        spmatrix = shortest_path(A.A, method='D', directed=False)
        self.coords = map.fit_transform(spmatrix).T
        self.reducer_ = map


    def create_mvu_embedding_(self):
        mvu = MaximumVarianceUnfolding(equation="berkley", solver=cp.SCS, solver_tol=1e-2,
                                       eig_tol=1.0e-10, solver_iters=2500,
                                       warm_start=False, seed=None)

        self.coords = mvu.fit_transform(self.graph.data.T, self.dim,
                                                      self.graph.nn, dropout_rate=0)
        self.reducer_ = mvu

    def create_lle_embedding_(self):
        lle = LocallyLinearEmbedding(n_components=self.dim, n_neighbors=self.graph.nn)
        self.coords = lle.fit_transform(self.graph.data.T).T
        self.reducer_ = lle

    def create_hlle_embedding_(self):
        hlle = LocallyLinearEmbedding(n_components=self.dim,
                                      n_neighbors=self.graph.nn,
                                      method='hessian')
        self.coords = hlle.fit_transform(self.graph.data.T).T
        self.reducer_ = hlle

    def create_le_embedding_(self):
        A = self.graph.adj
        dim = self.dim
        n = self.graph.n

        DH = get_inv_sqrt_diag_matrix(A)
        P = self.graph.get_matrix('trans')

        start_v = np.ones(n)
        # LR mode is much more stable, this is why we use P matrix largest eigenvalues
        eigvals, eigvecs = eigs(P, k=dim + 1, which='LR', v0=start_v, maxiter=n * 1000)
        # eigvals, vecs = eigs(nL, k = dim2 + 1, which = 'SM')

        eigvals = np.asarray([np.round(np.real(x), 6) for x in eigvals])

        if np.count_nonzero(eigvals == 1.0) > 1:
            raise Exception('Graph is not connected, LE will result in errors!')
        else:
            vecs = np.real(eigvecs.T[1:])
            vec_norms = np.array([np.real(sum([x * x for x in v])) for v in vecs])
            vecs = vecs / vec_norms[:, np.newaxis]
            vecs = vecs.dot(DH.toarray())
            self.coords = vecs

    def create_auto_le_embedding_(self):
        A = self.graph.adj
        dim = self.dim

        A = A.asfptype()
        vecs = spectral_embedding(A.A, n_components=dim, eigen_solver=None,
                                  random_state=None, eigen_tol=0.0, norm_laplacian=True, drop_first=True).T

        self.coords = vecs

    def create_dmaps_embedding_(self):
        raise Exception('not so easy to implement properly (https://sci-hub.se/10.1016/j.acha.2015.01.001)')

    def create_auto_dmaps_embedding_(self):
        dim = self.dim
        nn = self.graph.nn
        metric = self.graph.metric
        metric_args = self.graph.metric_args
        alpha = self.dm_alpha if hasattr(self, 'dm_alpha') else 1

        mydmap = dm.DiffusionMap.from_sklearn(n_evecs=dim,
                                              k=nn,
                                              epsilon='bgh',
                                              metric=metric,
                                              metric_params=metric_args,
                                              alpha=alpha)

        dmap = mydmap.fit_transform(self.init_data.T)

        self.coords = dmap.T
        self.reducer_ = dmap

    def create_tsne_embedding_(self):
        model = TSNE(n_components=self.dim, verbose=1)
        self.coords = model.fit_transform(self.init_data.T).T
        self.reducer_ = model

    def create_umap_embedding_(self):
        min_dist = self.min_dist
        reducer = umap.UMAP(n_neighbors=self.graph.nn, n_components=self.dim,
                            min_dist=min_dist)

        self.coords = reducer.fit_transform(self.graph.data.T).T
        self.reducer_ = reducer

    def create_ae_embedding_(self, continue_learning=0, epochs=50, lr=1e-3, seed=42, batch_size=32,
                             enc_kwargs=None, dec_kwargs=None, feature_dropout=0.2, train_size=0.8,
                             inter_dim=100, verbose=True, corr_hyperweight=0, mi_hyperweight=0,
                             minimize_mi_data=None, log_every=1, device=None):

        # ---------------------------------------------------------------------------

        torch.manual_seed(seed)
        torch.backends.cudnn.benchmark = False
        torch.backends.cudnn.deterministic = True

        # TODO: add train_test_split
        train_dataset = NeuroDataset(self.init_data[:, :int(train_size * self.init_data.shape[1])])
        test_dataset = NeuroDataset(self.init_data[:, int(train_size * self.init_data.shape[1]):])
        train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True)
        test_loader = DataLoader(test_dataset, batch_size=batch_size, shuffle=True)

        # ---------------------------------------------------------------------------
        if device is None:
            #  use gpu if available
            device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
            if verbose:
                print('device:', device)

        if not continue_learning:
            # create a model from `AE` autoencoder class
            # load it to the specified device, either gpu or cpu
            model = AE(orig_dim=self.init_data.shape[0], inter_dim=inter_dim, code_dim=self.dim,
                       enc_kwargs=enc_kwargs, dec_kwargs=dec_kwargs, device=device)

            model = model.to(device)
        else:
            model = self.nnmodel

        # create an optimizer object
        optimizer = optim.Adam(model.parameters(), lr=lr)

        # mean-squared error loss
        criterion = nn.MSELoss()

        def correlation_loss(data):
            #print('corr')
            corr = torch.corrcoef(data)
            #print(corr)
            nv = corr.shape[0]
            closs = torch.abs((torch.sum(torch.abs(corr)) - 1*nv)/(nv**2 - nv))  # average pairwise correlation amplitude
            #print(closs)
            return closs

        def data_orthogonality_loss(data, ortdata):

            # punishes big amplitude correlation coefficients between all variables from data and ortdata.
            # ortdata is supposed fixed
            # temporal workaround instead of MINE MI estimation

            # print('ortho')
            #print(ortdata)
            n1, n2 = data.shape[0], ortdata.shape[1]
            fulldata = torch.cat((data, ortdata), dim=0)
            corr = torch.corrcoef(fulldata)
            #print(corr)
            #print(corr[n1:, :n1])
            nvar = n1*n2
            closs = torch.abs(
                (torch.sum(torch.abs(corr))) / nvar)  # average pairwise correlation amplitude
            #print(closs)
            #print()
            return closs

        # ---------------------------------------------------------------------------
        f_dropout = nn.Dropout(feature_dropout)

        best_test_epoch = -1
        best_test_loss = 1e10
        best_test_model = None

        for epoch in range(epochs):
            loss = 0
            for batch_features, _, indices in train_loader:
                batch_features = batch_features.to(device)
                # reset the gradients back to zero
                # PyTorch accumulates gradients on subsequent backward passes
                optimizer.zero_grad()

                # compute reconstructions
                noisy_batch_features = f_dropout(torch.ones(batch_features.shape).to(device)) * batch_features
                outputs = model(noisy_batch_features.float())
                code = model.encoder(noisy_batch_features.float()).T

                '''
                # ==================== MINE experiment ========================

                from torch_mist.estimators import mine
                from torch_mist.utils.train import train_mi_estimator
                from torch_mist.utils import evaluate_mi

                minimize_mi_data = torch.tensor(minimize_mi_data[:, _]).float().to(device)
                print(minimize_mi_data.shape, code.shape)

                estimator = mine(
                    x_dim=minimize_mi_data.shape[0],
                    y_dim=code.shape[0],
                    hidden_dims=[64, 32],
                )


                # Train it on the given samples
                train_log = train_mi_estimator(
                    estimator=estimator,
                    train_data=(minimize_mi_data.T, code.T),
                    batch_size=batch_size,
                    max_iterations=1000,
                    device=device,
                    fast_train=True,
                    max_epochs=10
                )

                # Evaluate the estimator on the entirety of the data
                estimated_mi = evaluate_mi(
                    estimator=estimator,
                    data=(minimize_mi_data.T.to(device), code.T.to(device)),
                    batch_size=batch_size
                )

                print(f"Mutual information estimated value: {estimated_mi} nats")
                # ==================== MINE experiment ========================
                '''

                ortdata = torch.tensor(minimize_mi_data[:, indices]).float().to(device)
                # compute training reconstruction loss
                train_loss = criterion(outputs, batch_features.float()) + \
                                       corr_hyperweight * correlation_loss(code) + \
                                       mi_hyperweight * data_orthogonality_loss(code, ortdata)

                # compute accumulated gradients
                train_loss.backward()

                # perform parameter update based on current gradients
                optimizer.step()

                # add the mini-batch training loss to epoch loss
                loss += train_loss.item()

            # compute the epoch training loss
            loss = loss / len(train_loader)

            # display the epoch training loss
            if (epoch + 1) % log_every == 0:

                # compute loss on test part
                tloss = 0
                for batch_features, _, indices in test_loader:
                    batch_features = batch_features.to(device)
                    # compute reconstructions
                    noisy_batch_features = f_dropout(torch.ones(batch_features.shape).to(device)) * batch_features
                    outputs = model(noisy_batch_features.float())
                    code = model.encoder(noisy_batch_features.float()).T
                    ortdata = torch.tensor(minimize_mi_data[:, indices]).float().to(device)
                    # compute test reconstruction loss
                    test_loss = criterion(outputs, batch_features.float()) + \
                                       corr_hyperweight * correlation_loss(code) + \
                                       mi_hyperweight * data_orthogonality_loss(code, ortdata)
                    tloss += test_loss.item()

                # compute the epoch test loss
                tloss = tloss / len(test_loader)
                if tloss < best_test_loss:
                    best_test_loss = tloss
                    best_test_epoch = epoch + 1
                    best_test_model = model
                if verbose:
                    print(f"epoch : {epoch + 1}/{epochs}, train loss = {loss:.8f}, test loss = {tloss:.8f}")

        if verbose:
            if best_test_epoch != epochs + 1:
                print(f'best model: epoch {best_test_epoch}')

        self.nnmodel = best_test_model
        input_ = torch.tensor(self.init_data.T).float().to(device)
        self.coords = model.get_code_embedding(input_)
        self.nn_loss = tloss

    # -------------------------------------

    def create_vae_embedding_(self, continue_learning=0, epochs=50, lr=1e-3, seed=42, batch_size=32,
                              enc_kwargs=None, dec_kwargs=None, feature_dropout=0.2, kld_weight=1,
                              train_size=0.8, inter_dim=128, verbose=True):

    # TODO: add best model mechanism as above
        # ---------------------------------------------------------------------------
        torch.manual_seed(seed)
        torch.backends.cudnn.benchmark = False
        torch.backends.cudnn.deterministic = True

        # TODO: add train_test_split
        # TODO: move out data loading for autoencoders
        train_dataset = NeuroDataset(self.init_data[:, :int(train_size * self.init_data.shape[1])])
        test_dataset = NeuroDataset(self.init_data[:, int(train_size * self.init_data.shape[1]):])
        train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True)
        test_loader = DataLoader(test_dataset, batch_size=batch_size, shuffle=True)

        # ---------------------------------------------------------------------------
        #  use gpu if available
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        if not continue_learning:
            # create a model from `VAE` autoencoder class
            # load it to the specified device, either gpu or cpu
            model = VAE(orig_dim=len(self.init_data), inter_dim=inter_dim, code_dim=self.dim,
                        enc_kwargs=enc_kwargs, dec_kwargs=dec_kwargs)
            model = model.to(device)
        else:
            model = self.nnmodel

        # create an optimizer object
        # Adam optimizer with learning rate lr
        optimizer = optim.Adam(model.parameters(), lr=lr)

        # BCE error loss
        #criterion = nn.BCELoss(reduction='sum')
        criterion = nn.MSELoss()

        # ---------------------------------------------------------------------------
        f_dropout = nn.Dropout(feature_dropout)

        for epoch in range(epochs):
            loss = 0
            loss1 = 0
            loss2 = 0
            for batch_features, _ in train_loader:
                batch_features = batch_features.to(device)
                # reset the gradients back to zero
                # PyTorch accumulates gradients on subsequent backward passes
                optimizer.zero_grad()

                # compute reconstructions
                data = f_dropout(torch.ones(batch_features.shape).to(device)) * batch_features
                data = data.to(device)
                reconstruction, mu, logvar = model(data)

                # compute training reconstruction loss
                mse_loss = criterion(reconstruction, data)
                kld_loss = -0.5 * torch.sum(
                    1 + logvar - mu.pow(2) - logvar.exp())  # * train_dataset.__len__()/batch_size
                train_loss = mse_loss + kld_weight*kld_loss

                # compute accumulated gradients
                train_loss.backward()

                # perform parameter update based on current gradients
                optimizer.step()

                # add the mini-batch training loss to epoch loss
                loss += train_loss.item()
                loss1 += mse_loss.item()
                loss2 += kld_loss.item()

            # compute the epoch training loss
            loss = loss / len(train_loader)
            loss1 = loss1 / len(train_loader)
            loss2 = loss2 / len(train_loader)

            # display the epoch training loss
            # display the epoch training loss
            if (epoch + 1) % 10 == 0:
                # compute loss on test part
                tloss = 0
                for batch_features, _ in test_loader:
                    data = f_dropout(torch.ones(batch_features.shape).to(device)) * batch_features
                    data = data.to(device)
                    reconstruction, mu, logvar = model(data)

                    # compute training reconstruction loss
                    mse_loss = criterion(reconstruction, data)
                    kld_loss = -0.5 * torch.sum(
                        1 + logvar - mu.pow(2) - logvar.exp())  # * train_dataset.__len__()/batch_size
                    test_loss = mse_loss + kld_weight * kld_loss
                    tloss += test_loss.item()

                # compute the epoch training loss
                tloss = tloss / len(test_loader)
                if verbose:
                    print(f"epoch : {epoch + 1}/{epochs}, train loss = {loss:.8f}, test loss = {tloss:.8f}")

        self.nnmodel = model
        input_ = torch.tensor(self.init_data.T).float().to(device)
        self.coords = model.get_code_embedding(input_)

        # -------------------------------------

    def continue_learning(self, add_epochs, kwargs={}):
        if self.all_params['e_method_name'] not in ['ae', 'vae']:
            raise Exception('This is not a DL-based method!')

        fn = getattr(self, 'create_' + self.all_params['e_method_name'] + '_embedding_')
        fn(continue_learning=1, epochs=add_epochs, kwargs=kwargs)




import os
from os.path import join, splitext
import tqdm
import matplotlib.pyplot as plt

from ssqueezepy import cwt
from ssqueezepy.wavelets import Wavelet, time_resolution

from scipy.ndimage.filters import gaussian_filter1d
from scipy.signal import argrelmax
from numba import njit

from .wavelet_ridge import *

WVT_EVENT_DETECTION_PARAMS = {
                  'fps': 20,        # fps, frames
                  'sigma': 8,      # smoothing parameter for peak detection, frames
                  'beta': 2,       # Generalized Morse Wavelet parameter, FIXED
                  'gamma': 3,      # Generalized Morse Wavelet parameter, FIXED
                  'eps': 10,         # spacing between consecutive events, frames
                  'manual_scales': np.logspace(2.5,5.5,50, base=2),

                  # ridge filtering params
                  'scale_length_thr': 40,  # min number of scales where ridge is present thr, higher = less events. max=len(manual_scales)
                  'max_scale_thr': 7,      # index of a scale with max ridge intensity thr, higher = less events. < 5 = noise, > 20 = huge events
                  'max_ampl_thr': 0.05,    # max ridge intensity thr, higher = less events. < 5 = noise, > 20 = huge events
                  'max_dur_thr': 200,      # max event duration thr, higher = more events (but probably strange ones)
}
MIN_EVENT_DUR = 0.5 # sec
MAX_EVENT_DUR = 2.5 # sec


def wvt_viz(x, Wx):
    fig, axs = plt.subplots(2, 1, figsize=(12,12))
    axs[0].set_xlim(0, len(x))
    axs[0].plot(x, c='b')
    axs[1].imshow(np.abs(Wx), aspect='auto', cmap='turbo')


def get_cwt_ridges(sig, wavelet=None, fps=20, scmin=150, scmax=250, all_wvt_times=None, wvt_scales=None):
    if wvt_scales is not None:
        scales = wvt_scales
    else:
        scales = 'log-piecewise'
    W, wvt_scales = cwt(sig, wavelet=wavelet, fs=fps, scales=scales)

    #wvtdata = np.real(np.abs(W))
    wvtdata = np.real(W)
    scale_inds = np.arange(scmin, scmax)[::-1]
    if all_wvt_times is None:
        all_wvt_times = [time_resolution(wavelet, scale=wvt_scales[sc], nondim=False, min_decay=200) for sc in scale_inds]

    # determine peak positions for all scales
    peaks = np.zeros((len(scale_inds), len(sig)))

    all_ridges = []
    for i, si in enumerate(scale_inds[:]):
        wvt_time = all_wvt_times[i]
        max_inds = argrelmax(wvtdata[si,:], order=10)[0]
        peaks[i, max_inds] = wvtdata[si, max_inds]
        #max_inds = np.nonzero(peaks[i,:])[0]
        #print(peaks[i, max_inds])

        if len(all_ridges) == 0:
            all_ridges = [Ridge(mi, peaks[i, mi], wvt_scales[si], wvt_time) for mi in max_inds]
        else:
            # 1. extend old ridges
            prev_wvt_time = all_wvt_times[i-1]
            live_ridges = [ridge for ridge in all_ridges if not ridge.terminated]
            maxima_used_for_prolongation = []

            for ridge in live_ridges:
                # 1.1 get ridge tip from previous scale
                last_max_index = ridge.tip()
                # 1.2 compute time window based on 68% of wavelet energy
                wlb, wrb = last_max_index - prev_wvt_time, last_max_index + prev_wvt_time
                # 1.3 get list of candidate maxima of the current scale falling into the window
                candidates = [mi for mi in max_inds if (mi > wlb) and (mi < wrb)]
                # 1.4 extending ridges
                if len(candidates) == 0:
                    # gaps lead to ridge termination
                    #print(f'ridge with start time {ridge.indices[0]} terminated')
                    ridge.terminate()
                elif len(candidates) == 1:
                    # extend ridge
                    cand = candidates[0]
                    ridge.extend(cand, peaks[i, cand], wvt_scales[si], wvt_time)
                    maxima_used_for_prolongation.append(cand)
                    #print(f'ridge with start time {ridge.indices[0]} extended')
                else:
                    # extend ridge with the best maximum, others will later form new ridges
                    best_cand = candidates[np.argmax(peaks[i, np.array(candidates)])]
                    ridge.extend(best_cand, peaks[i, best_cand], wvt_scales[si], wvt_time)
                    maxima_used_for_prolongation.append(best_cand)
                    #maxima_used_for_prolongation.extend(candidates)

            # 2. generate new ridges
            new_ridges = [Ridge(mi, peaks[i, mi], wvt_scales[si], wvt_time) for mi in max_inds if mi not in maxima_used_for_prolongation]
            all_ridges.extend(new_ridges)

    for r in all_ridges:
        r.terminate()

    return all_ridges


#@njit()
def get_cwt_ridges_fast(wvtdata, peaks, wvt_times, wvt_scales):
    # determine peak positions for all scales

    start = True
    for si in range(wvtdata.shape[0]):
        wvt_time = wvt_times[si]
        max_inds = np.nonzero(peaks[si,:])[0]

        if start:
            all_ridges = [Ridge(mi, peaks[si, mi], wvt_scales[si], wvt_time) for mi in max_inds]
            start = False
        else:
            # 1. extend old ridges
            prev_wvt_time = wvt_times[si-1]
            live_ridges = [ridge for ridge in all_ridges if not ridge.terminated]
            maxima_used_for_prolongation = []

            for ridge in live_ridges:
                # 1.1 get ridge tip from previous scale
                last_max_index = ridge.tip()
                # 1.2 compute time window based on 68% of wavelet energy
                wlb, wrb = last_max_index - prev_wvt_time, last_max_index + prev_wvt_time
                # 1.3 get list of candidate maxima of the current scale falling into the window
                candidates = [mi for mi in max_inds if (mi > wlb) and (mi < wrb)]
                # 1.4 extending ridges
                if len(candidates) == 0:
                    # gaps lead to ridge termination
                    #print(f'ridge with start time {ridge.indices[0]} terminated')
                    ridge.terminate()
                elif len(candidates) == 1:
                    # extend ridge
                    cand = candidates[0]
                    ridge.extend(cand, peaks[si, cand], wvt_scales[si], wvt_time)
                    maxima_used_for_prolongation.append(cand)
                    #print(f'ridge with start time {ridge.indices[0]} extended')
                else:
                    # extend ridge with the best maximum, others will later form new ridges
                    best_cand = candidates[np.argmax(peaks[si, np.array(candidates)])]
                    ridge.extend(best_cand, peaks[si, best_cand], wvt_scales[si], wvt_time)
                    maxima_used_for_prolongation.append(best_cand)
                    #maxima_used_for_prolongation.extend(candidates)

            # 2. generate new ridges
            new_ridges = [Ridge(mi, peaks[si, mi], wvt_scales[si], wvt_time) for mi in max_inds if mi not in maxima_used_for_prolongation]
            all_ridges.extend(new_ridges)

    for r in all_ridges:
        r.terminate()

    return all_ridges


def passing_criterion(ridge, scale_length_thr=40, max_scale_thr=10, max_ampl_thr=0.05, max_dur_thr=100):
    crit = ridge.length >= scale_length_thr and ridge.max_scale >= max_scale_thr and ridge.max_ampl >= max_ampl_thr and ridge.duration <= max_dur_thr
    return crit


def get_events_from_ridges(all_ridges, scale_length_thr=40, max_scale_thr=10, max_ampl_thr=0.05, max_dur_thr=100):
    event_ridges = [r for r in all_ridges if passing_criterion(r,
                                                               scale_length_thr=scale_length_thr,
                                                               max_scale_thr=max_scale_thr,
                                                               max_ampl_thr=max_ampl_thr,
                                                               max_dur_thr=max_dur_thr)]

    st_evinds = [r.indices[0] for r in event_ridges]
    end_evinds = [r.indices[-1] for r in event_ridges]
    return st_evinds, end_evinds


def events_from_trace(trace, wavelet, manual_scales, rel_wvt_times,
                      fps=20, sigma=8, eps=10,
                      scale_length_thr=40,
                      max_scale_thr=7,
                      max_ampl_thr=0.05,
                      max_dur_thr=200):

    trace = (trace - min(trace))/(max(trace) - min(trace))
    sig = gaussian_filter1d(trace, sigma=sigma)

    W, wvt_scales = cwt(sig, wavelet=wavelet, fs=fps, scales=manual_scales)
    rev_wvtdata = np.real(W)

    all_max_inds = argrelmax(rev_wvtdata, axis=1, order=eps)
    peaks = np.zeros(rev_wvtdata.shape)
    peaks[all_max_inds] = rev_wvtdata[all_max_inds]

    all_ridges = get_cwt_ridges_fast(rev_wvtdata, peaks, rel_wvt_times, manual_scales)

    st_evinds, end_evinds = get_events_from_ridges(all_ridges,
                                                   scale_length_thr=scale_length_thr,
                                                   max_scale_thr=max_scale_thr,
                                                   max_ampl_thr=max_ampl_thr,
                                                   max_dur_thr=max_dur_thr)

    return all_ridges, st_evinds, end_evinds


def extract_wvt_events(traces, wvt_kwargs):
    fps = wvt_kwargs.get('fps', 20)
    beta = wvt_kwargs.get('beta', 2)
    gamma = wvt_kwargs.get('gamma', 3)
    sigma = wvt_kwargs.get('sigma', 8)
    eps = wvt_kwargs.get('eps', 10)
    manual_scales = wvt_kwargs.get('manual_scales', np.logspace(2.5,5.5,50, base=2))

    scale_length_thr = wvt_kwargs.get('scale_length_thr', 40)
    max_scale_thr = wvt_kwargs.get('max_scale_thr', 7)
    max_ampl_thr = wvt_kwargs.get('max_ampl_thr', 0.05)
    max_dur_thr = wvt_kwargs.get('max_dur_thr', 200)

    wavelet = Wavelet(('gmw', {'gamma': gamma, 'beta': beta, 'centered_scale': True}), N=8196)

    rel_wvt_times = [time_resolution(wavelet, scale=sc, nondim=False, min_decay=200) for sc in manual_scales]

    st_ev_inds = []
    end_ev_inds = []
    all_ridges = []
    for i, trace in tqdm.tqdm(enumerate(traces), total=len(traces)):
        ridges, st_ev, end_ev = events_from_trace(trace,
                                                  wavelet,
                                                  manual_scales,
                                                  rel_wvt_times,
                                                  fps=fps,
                                                  sigma=sigma,
                                                  eps=eps,
                                                  scale_length_thr=scale_length_thr,
                                                  max_scale_thr=max_scale_thr,
                                                  max_ampl_thr=max_ampl_thr,
                                                  max_dur_thr=max_dur_thr)

        st_ev_inds.append(st_ev)
        end_ev_inds.append(end_ev)
        all_ridges.append(ridges)

    return st_ev_inds, end_ev_inds, all_ridges


def events_to_ts_array(length, st_ev_inds, end_ev_inds, fps):
    ncells = len(end_ev_inds)
    spikes = np.zeros((ncells, length))

    mindur = int(MIN_EVENT_DUR * fps)
    maxdur = int(MAX_EVENT_DUR * fps)

    for i in range(ncells):
        for j in range(len(st_ev_inds[i])):
            start = int(st_ev_inds[i][j])
            end = int(end_ev_inds[i][j])
            start_, end_ = min(start, end), max(start, end)
            dur = end_ - start_
            if mindur <= dur <= maxdur:
                spikes[i, start_: end_] = 1
            elif dur > maxdur:
                spikes[i, start_: start_ + maxdur] = 1
            else:
                middle = (start_ + end_)//2
                spikes[i, int(middle - mindur//2): int(middle + mindur//2)] = 1

    return spikes



from pynndescent.distances import named_distances
import sys

class DRMethod(object):

    def __init__(self, is_linear, requires_graph, requires_distmat, nn_based):
        self.is_linear = is_linear
        self.requires_graph = requires_graph
        self.requires_distmat = requires_distmat
        self.nn_based = nn_based


METHODS_DICT = {
    'pca': DRMethod(1, 0, 0, 0),
    'le': DRMethod(0, 1, 0, 0),
    'auto_le': DRMethod(0, 1, 0, 0),
    'dmaps': DRMethod(0, 1, 0, 0),
    'auto_dmaps': DRMethod(0, 1, 0, 0),
    'mds': DRMethod(0, 0, 1, 0),
    'isomap': DRMethod(0, 1, 0, 0),
    'lle': DRMethod(0, 1, 0, 0),
    'hlle': DRMethod(0, 1, 0, 0),
    'mvu': DRMethod(0, 1, 0, 0),
    'ae': DRMethod(0, 0, 0, 1),
    'vae': DRMethod(0, 0, 0, 1),
    'tsne': DRMethod(0, 0, 0, 0),
    'umap': DRMethod(0, 1, 0, 0)
}

GRAPH_CONSTRUCTION_METHODS = ['knn', 'auto_knn', 'eps', 'eknn', 'umap', 'tsne']

EMBEDDING_CONSTRUCTION_METHODS = ['pca',
                                  'le',
                                  'auto_le',
                                  'dmaps',
                                  'auto_dmaps',
                                  'mds',
                                  'isomap',
                                  'lle',
                                  'hlle',
                                  'mvu',
                                  'ae',
                                  'vae',
                                  'tsne',
                                  'umap']

# TODO: implement random projections


def m_param_filter(para):
    '''
    This function prunes parameters that are excessive for
    chosen distance matrix construction method
    '''
    name = para['metric_name']
    appr_keys = ['metric_name']

    if not (para['sigma'] is None):
        appr_keys.append('sigma')

    if name not in named_distances:
        if name == 'hyperbolic':
            # para['metric_name'] = globals()[name]
            pass
        else:
            raise Exception('this custom metric is not implemented!')

    if name == 'minkowski':
        appr_keys.append('p')

    return {key: para[key] for key in appr_keys}


def g_param_filter(para):
    '''
    This function prunes parameters that are excessive for
    chosen graph construction method
    '''
    gmethod = para['g_method_name']
    appr_keys = ['g_method_name', 'max_deleted_nodes', 'weighted', 'dist_to_aff']

    if gmethod in ['knn', 'auto_knn', 'umap']:
        appr_keys.extend(['nn'])

    elif gmethod == 'eps':
        appr_keys.extend(['eps', 'eps_min'])

    elif gmethod == 'eknn':
        appr_keys.extend(['eps', 'eps_min', 'nn'])

    elif gmethod == 'tsne':
        appr_keys.extend(['perplexity'])

    return {key: para[key] for key in appr_keys}


def e_param_filter(para):
    '''
    This function prunes parameters that are excessive for the
    chosen embedding construction method
    '''

    appr_keys = ['e_method', 'e_method_name', 'dim']

    if para['e_method_name'] == 'umap':
        appr_keys.append('min_dist')

    if para['e_method_name'] in ['dmaps', 'auto_dmaps']:
        appr_keys.append('dm_alpha')

    return {key: para[key] for key in appr_keys}


from io import StringIO
import sys

class Capturing(list):
    def __enter__(self):
        self._stdout = sys.stdout
        sys.stdout = self._stringio = StringIO()
        return self
    def __exit__(self, *args):
        self.extend(self._stringio.getvalue().splitlines())
        del self._stringio    # free up some memory
        sys.stdout = self._stdout


def show_output(output):
    if len(output) == 0:
        print('log is empty')
    else:
        for s in output:
            print(s)



from sklearn.datasets import make_swiss_roll
import numpy as np
from src.driada.dim_reduction.data import *

n_gaussian = 10
n_swiss_roll = 1000

def create_default_data(n=n_gaussian, T=10000):
    C = np.zeros((n,n))
    C[1, 9] = 0.9
    C[2, 8] = 0.8
    C = (C + C.T)
    np.fill_diagonal(C, 1)
    signals = np.random.multivariate_normal(np.zeros(n),
                                            C,
                                            size=T,
                                            check_valid='raise').T
    return signals


def create_swiss_roll_data():
    random_state = 42
    data, _ = make_swiss_roll(n_samples=n_swiss_roll,
                              noise=0.0,
                              random_state=random_state,
                              hole=False)
    return data.T


def test_corrmat():
    data = create_default_data()
    D = MVData(data)
    cm = D.corr_mat()
    assert np.allclose(np.diag(cm), np.ones(n_gaussian))


def test_proximity_graph():
    data = create_swiss_roll_data()

    metric_params = {
        'metric_name': 'l2',
        'sigma': 1,
        'p': 2
    }

    graph_params = {
        'g_method_name': 'knn',
        'weighted': 0,
        'nn': 10,
        'max_deleted_nodes': 0.2,
        'dist_to_aff': 'hk'
    }

    D = MVData(data)
    G = D.get_proximity_graph(metric_params, graph_params)


def test_pca():
    data = create_swiss_roll_data()
    D = MVData(data)

    embedding_params = {
        'e_method_name': 'pca',
        'dim': 2
    }

    embedding_params['e_method'] = METHODS_DICT[embedding_params['e_method_name']]
    emb = D.get_embedding(embedding_params)
    assert emb.coords.shape == (2, n_swiss_roll)
    assert type(emb.coords[0,0]) in [np.float64, np.float32]


def test_le():
    data = create_swiss_roll_data()
    D = MVData(data)

    metric_params = {
        'metric_name': 'l2',
        'sigma': 1,
        'p': 2
    }

    graph_params = {
        'g_method_name': 'knn',
        'weighted': 0,
        'nn': 10,
        'max_deleted_nodes': 0.2,
        'dist_to_aff': 'hk'
    }

    embedding_params = {
        'e_method_name': 'le',
        'dim': 2
    }

    embedding_params['e_method'] = METHODS_DICT[embedding_params['e_method_name']]
    emb = D.get_embedding(embedding_params, g_params=graph_params, m_params=metric_params)
    assert emb.coords.shape == (2, n_swiss_roll)
    assert type(emb.coords[0, 0]) in [np.float64, np.float32]


def test_auto_le():
    data = create_swiss_roll_data()
    D = MVData(data)

    metric_params = {
        'metric_name': 'l2',
        'sigma': 1,
        'p': 2
    }

    graph_params = {
        'g_method_name': 'knn',
        'weighted': 0,
        'nn': 10,
        'max_deleted_nodes': 0.2,
        'dist_to_aff': 'hk'
    }

    embedding_params = {
        'e_method_name': 'auto_le',
        'dim': 2
    }

    embedding_params['e_method'] = METHODS_DICT[embedding_params['e_method_name']]
    emb = D.get_embedding(embedding_params, g_params=graph_params, m_params=metric_params)
    assert emb.coords.shape == (2, n_swiss_roll)
    assert type(emb.coords[0, 0]) in [np.float64, np.float32]


def test_umap():
    data = create_swiss_roll_data()
    D = MVData(data)

    metric_params = {
        'metric_name': 'l2',
        'sigma': 1,
        'p': 2
    }

    graph_params = {
        'g_method_name': 'knn',
        'weighted': 0,
        'nn': 10,
        'max_deleted_nodes': 0.2,
        'dist_to_aff': 'hk'
    }

    embedding_params = {
        'e_method_name': 'umap',
        'dim': 2,
        'min_dist': 0.1
    }

    embedding_params['e_method'] = METHODS_DICT[embedding_params['e_method_name']]
    emb = D.get_embedding(embedding_params, g_params=graph_params, m_params=metric_params)
    assert emb.coords.shape == (2, n_swiss_roll)
    assert type(emb.coords[0, 0]) in [np.float64, np.float32]


def test_isomap():
    data = create_swiss_roll_data()
    D = MVData(data)

    metric_params = {
        'metric_name': 'l2',
        'sigma': 1,
        'p': 2
    }

    graph_params = {
        'g_method_name': 'knn',
        'weighted': 0,
        'nn': 10,
        'max_deleted_nodes': 0.2,
        'dist_to_aff': 'hk'
    }

    embedding_params = {
        'e_method_name': 'isomap',
        'dim': 2,
        'min_dist': 0.1
    }

    embedding_params['e_method'] = METHODS_DICT[embedding_params['e_method_name']]
    emb = D.get_embedding(embedding_params, g_params=graph_params, m_params=metric_params)
    assert emb.coords.shape == (2, n_swiss_roll)
    assert type(emb.coords[0, 0]) in [np.float64, np.float32]


def test_tsne():
    data = create_swiss_roll_data()
    D = MVData(data)

    embedding_params = {
        'e_method_name': 'tsne',
        'dim': 2,
    }

    embedding_params['e_method'] = METHODS_DICT[embedding_params['e_method_name']]
    emb = D.get_embedding(embedding_params)
    assert emb.coords.shape == (2, n_swiss_roll)
    assert type(emb.coords[0, 0]) in [np.float64, np.float32]


def test_auto_dmaps():
    data = create_swiss_roll_data()
    D = MVData(data)

    metric_params = {
        'metric_name': 'l2',
        'sigma': 1,
        'p': 2
    }

    graph_params = {
        'g_method_name': 'knn',
        'weighted': 0,
        'nn': 10,
        'max_deleted_nodes': 0.2,
        'dist_to_aff': 'hk'
    }

    embedding_params = {
        'e_method_name': 'auto_dmaps',
        'dim': 2,
        'dm_alpha': 1
    }

    embedding_params['e_method'] = METHODS_DICT[embedding_params['e_method_name']]
    emb = D.get_embedding(embedding_params, g_params=graph_params, m_params=metric_params)
    assert emb.coords.shape == (2, n_swiss_roll)
    assert type(emb.coords[0, 0]) in [np.float64, np.float32]




#from .gdrive.download import *
#from .gdrive.upload import *

# TODO: include environment.yml
# TODO: unit tests for calculations (https://goodresearch.dev/testing)
# TODO: document everything in Google Style (https://goodresearch.dev/docs)
# TODO: publish docs on Readthedocs (https://docs.readthedocs.io/en/stable/tutorial/index.html#importing-the-project-to-read-the-docs_
# TODO: update ReadMe (https://goodresearch.dev/pipelines)

# TODO: probably add automatic scanning for every .py file (dangerous and not pythonic, but convenient)
'''
import glob
from os.path import dirname, basename, isfile, join
modules = glob.glob(join(dirname(__file__), "*", "*.py"), recursive=True)
__all__ = [basename(f)[:-3] for f in modules if isfile(f) and not basename(f).startswith('_')]
'''

# TODO: make Readme structured as a numbered list, like in mango


from .stats import *
from .intense_base import compute_mi_stats


def compute_cell_feat_mi_significance(exp,
                                      cell_bunch=None,
                                      feat_bunch=None,
                                      data_type='calcium',
                                      mode='two_stage',
                                      n_shuffles_stage1=100,
                                      n_shuffles_stage2=10000,
                                      joint_distr=False,
                                      mi_distr_type='gamma',
                                      noise_ampl=1e-3,
                                      ds=1,
                                      use_precomputed_stats=True,
                                      save_computed_stats=True,
                                      force_update=False,
                                      topk1=1,
                                      topk2=5,
                                      multicomp_correction='holm',
                                      pval_thr=0.01,
                                      find_optimal_delays=True,
                                      shift_window=5,
                                      verbose=True):

    """
    Calculates significant neuron-feature pairs

    Parameters
    ----------
    exp: Experiment instance
        Experiment object to read and write data from

    cell_bunch: int, iterable or None
        Neuron indices. By default, (cell_bunch=None), all neurons will be taken

    feat_bunch: str, iterable or None
        Feature names. By default, (feat_bunch=None), all single features will be taken

    data_type: str
        Data used for INTENS computations. Can be 'calcium' or 'spikes'

    mode: str
        Computation mode. 3 modes are available:
        'stage1': perform preliminary scanning with "n_shuffles_stage1" shuffles only.
                  Rejects strictly non-significant neuron-feature pairs, does not give definite results
                  about significance of the others.
        'stage2': skip stage 1 and perform full-scale scanning ("n_shuffles_stage2" shuffles) of all neuron-feature pairs.
                  Gives definite results, but can be very time-consuming. Also reduces statistical power
                  of multiple comparison tests, since the number of hypotheses is very high.
        'two_stage': prune non-significant pairs during stage 1 and perform thorough testing for the rest during stage 2.
                     Recommended mode.
        default: 'two-stage'

    n_shuffles_stage1: int
        number of shuffles for first stage
        default: 100

    n_shuffles_stage2: int
        number of shuffles for second stage
        default: 10000

    joint_distr: bool
        if joint_distr=True, ALL features in feat_bunch will be treated as components of a single multifeature
        For example, 'x' and 'y' features will be put together into ('x','y') multifeature.
        default: False

    mi_distr_type: str
        Distribution type for shuffles MI distribution fit. Supported options are "gamma" and "lognormal"
        default: "gamma"

    noise_ampl: float
        Small noise amplitude, which is added to MI and shuffled MI to improve numerical fit
        default: 1e-3

    ds: int
        Downsampling constant. Every "ds" point will be taken from the data time series.
        Reduces the computational load, but needs caution since with large "ds" some important information may be lost.
        Experiment class performs an internal check for this effect.
        default: 1

    use_precomputed_stats: bool
        Whether to use stats saved in Experiment instance. Stats are accumulated separately for stage1 and stage2.
        Notes on stats data rewriting (if save_computed_stats=True):
        If you want to recalculate stage1 results only, use "use_precomputed_stats=False" and "mode='stage1'".
        Stage 2 stats will be erased since they will become irrelevant.
        If you want to recalculate stage2 results only, use "use_precomputed_stats=True" and "mode='stage2'" or "mode='two-stage'"
        If you want to recalculate everything, use "use_precomputed_stats=False" and "mode='two-stage'"
        default: True

    save_computed_stats: bool
        Whether to save computed stats to Experiment instance
        default: True

    force_update: bool
        Whether to force saved statistics data update in case the collision between actual data hashes and
        saved stats data hashes is found (for example, if neuronal or behavior data has been changed externally).
        default: False

    topk1: int
        true MI for stage 1 should be among topk1 MI shuffles
        default: 1

    topk2: int
        true MI for stage 2 should be among topk2 MI shuffles
        default: 5

    multicomp_correction: str or None
        type of multiple comparisons correction. Supported types are None (no correction),
        "bonferroni" and "holm".
        default: 'holm'

    pval_thr: float
        pvalue threshold. if multicomp_correction=None, this is a p-value for a single pair.
        Otherwise it is a FWER significance level.

    find_optimal_delays: bool
        Allows slight shifting (not more than +- shift_window) of time series,
        selects a shift with the highest MI as default.
        default: True

    shift_window: int
        Window for optimal shift search (seconds). Optimal shift (in frames) will lie in the range
        -shift_window*fps <= opt_shift <= shift_window*fps
        Has no effect if find_optimal_delays = False

    Returns
    -------
    stats: dict of dict of dicts
        Outer dict: dynamic features, inner dict: cells, last dict: stats.
        Can be easily converted to pandas DataFrame by pd.DataFrame(stats)
    """

    exp.check_ds(ds)

    cell_ids = exp._process_cbunch(cell_bunch)
    feat_ids = exp._process_fbunch(feat_bunch, allow_multifeatures=False)
    cells = [exp.neurons[cell_id] for cell_id in cell_ids]

    if data_type == 'calcium':
        signals = [cell.ca for cell in cells]
    elif data_type == 'spikes':
        signals = [cell.sp for cell in cells]
    else:
        raise ValueError('"data_type" can be either "calcium" or "spikes"')
    #min_shifts = [int(cell.get_t_off() * MIN_CA_SHIFT) for cell in cells]
    feats = [exp.dynamic_features[feat_id] for feat_id in feat_ids]

    if joint_distr:
        feat_ids = [tuple(sorted(feat_ids))]

    n, t, f = len(cells), exp.n_frames, len(feat_ids)

    precomputed_mask_stage1 = np.ones((n,f))
    precomputed_mask_stage2 = np.ones((n,f))

    if use_precomputed_stats:
        print('Retrieving saved stats data...')
        # 0 in mask values means precomputed results are found, calculation will be skipped.
        # 1 in mask values means precomputed results are not found or incomplete, calculation will proceed.

        for i, cell_id in enumerate(cell_ids):
            for j, feat_id in enumerate(feat_ids):
                try:
                    # TODO: add data_type everywhere in stats
                    pair_stats = exp.get_neuron_feature_pair_stats(cell_id, feat_id)
                except ValueError:
                    if isinstance(feat_id, str):
                        raise ValueError(f'Unknown single feature in feat_bunch: {feat_id}. Check initial data')
                    else:
                        exp._add_multifeature_to_data_hashes(feat_id, mode=data_type)
                        exp._add_multifeature_to_stats(feat_id)
                        pair_stats = DEFAULT_STATS.copy()

                current_data_hash = exp._data_hashes[data_type][feat_id][cell_id]

                if stats_not_empty(pair_stats, current_data_hash, stage=1):
                    precomputed_mask_stage1[i,j] = 0
                if stats_not_empty(pair_stats, current_data_hash, stage=2):
                    precomputed_mask_stage2[i,j] = 0

    combined_precomputed_mask = np.ones((n, f))
    if mode in ['stage2', 'two_stage']:
        combined_precomputed_mask[np.where((precomputed_mask_stage1 == 0) & (precomputed_mask_stage2 == 0))] = 0
    elif mode == 'stage1':
        combined_precomputed_mask[np.where(precomputed_mask_stage1 == 0)] = 0
    else:
        raise ValueError('Wrong mode!')

    computed_stats, computed_significance, info = compute_mi_stats(signals,
                                                                   feats,
                                                                   mode=mode,
                                                                   names1=cell_ids,
                                                                   names2=feat_ids,
                                                                   precomputed_mask_stage1=precomputed_mask_stage1,
                                                                   precomputed_mask_stage2=precomputed_mask_stage2,
                                                                   n_shuffles_stage1=n_shuffles_stage1,
                                                                   n_shuffles_stage2=n_shuffles_stage2,
                                                                   joint_distr=joint_distr,
                                                                   mi_distr_type=mi_distr_type,
                                                                   noise_ampl=noise_ampl,
                                                                   ds=ds,
                                                                   topk1=topk1,
                                                                   topk2=topk2,
                                                                   multicomp_correction=multicomp_correction,
                                                                   pval_thr=pval_thr,
                                                                   find_optimal_delays=find_optimal_delays,
                                                                   shift_window=shift_window*exp.fps,
                                                                   verbose=verbose)

    exp.optimal_nf_delays = info['optimal_delays']
    # add hash data and update Experiment saved statistics and significance if needed
    for i, cell_id in enumerate(cell_ids):
        for j, feat_id in enumerate(feat_ids):
            # TODO: add check for non-existing feature if use_precomputed_stats==False
            computed_stats[cell_id][feat_id]['data_hash'] = exp._data_hashes[data_type][feat_id][cell_id]

            mi_val = computed_stats[cell_id][feat_id].get('mi')
            if mi_val is not None:
                feat_entropy = exp.get_feature_entropy(feat_id, ds=ds)
                ca_entropy = exp.neurons[int(cell_id)].ca.get_entropy(ds=ds)
                computed_stats[cell_id][feat_id]['rel_mi_beh'] = mi_val / feat_entropy
                computed_stats[cell_id][feat_id]['rel_mi_ca'] = mi_val / ca_entropy

            if save_computed_stats:
                stage2_only = True if mode == 'stage2' else False
                if combined_precomputed_mask[i,j]:
                    exp.update_neuron_feature_pair_stats(computed_stats[cell_id][feat_id],
                                                         cell_id,
                                                         feat_id,
                                                         force_update=force_update,
                                                         stage2_only=stage2_only)

                    sig = computed_significance[cell_id][feat_id]
                    exp.update_neuron_feature_pair_significance(sig, cell_id, feat_id)

    return computed_stats, computed_significance, info



from src.driada.intense.intense_base import compute_mi_stats
from src.driada.information.info_base import TimeSeries
from src.driada.utils.data import retrieve_relevant_from_nested_dict
import numpy as np


def create_correlated_ts(n=100, T=10000):
    np.random.seed(42)
    C = np.zeros((n,n))
    C[1, 99] = 0.9
    C[2, 98] = 0.8
    C[5, 95] = 0.7
    C = (C + C.T)
    np.fill_diagonal(C, 1)
    signals = np.random.multivariate_normal(np.zeros(n),
                                            C,
                                            size=T,
                                            check_valid='raise').T

    # cutting coherency windows, setting to 0 outside them
    w=100
    starts = np.random.choice(np.arange(w, T-w), size=10)
    nnz_time_inds = []
    for st in starts:
        nnz_time_inds.extend([st + _ for _ in range(w)])

    cropped_signals = np.zeros((n,T))
    cropped_signals[:, np.array(nnz_time_inds)] = signals[:, np.array(nnz_time_inds)]

    # add noise to remove coinciding values
    small_noise = np.random.multivariate_normal(np.zeros(n),
                                                np.eye(n),
                                                size=T,
                                                check_valid='raise').T*0.2

    cropped_signals += small_noise

    tslist1 = [TimeSeries(sig, discrete=False) for sig in cropped_signals[:n // 2,:]]
    tslist2 = [TimeSeries(sig, discrete=False) for sig in cropped_signals[n // 2:, :]]

    return tslist1, tslist2


def test_stage1():
    tslist1, tslist2 = create_correlated_ts()
    for ts in tslist1:
        ts.shuffle_mask[:50] = 0
    for ts in tslist2:
        ts.shuffle_mask[:50] = 0
    computed_stats, computed_significance, info = compute_mi_stats(tslist1,
                                                             tslist2,
                                                             mode='stage1',
                                                             n_shuffles_stage1=100,
                                                             joint_distr=False,
                                                             mi_distr_type='gamma',
                                                             noise_ampl=1e-3,
                                                             ds=1,
                                                             topk1=1,
                                                             verbose=True)

    rel_stats_pairs = retrieve_relevant_from_nested_dict(computed_stats, 'pre_rval', 1)
    rel_sig_pairs = retrieve_relevant_from_nested_dict(computed_significance, 'stage1', True)
    assert rel_sig_pairs == rel_stats_pairs


def test_two_stage():
    tslist1, tslist2 = create_correlated_ts()
    for ts in tslist1:
        ts.shuffle_mask[:50] = 0
    for ts in tslist2:
        ts.shuffle_mask[:50] = 0
    computed_stats, computed_significance, info = compute_mi_stats(tslist1,
                                                                     tslist2,
                                                                     mode='two_stage',
                                                                     n_shuffles_stage1=100,
                                                                     n_shuffles_stage2=1000,
                                                                     joint_distr=False,
                                                                     mi_distr_type='gamma',
                                                                     noise_ampl=1e-3,
                                                                     ds=1,
                                                                     topk1=1,
                                                                     topk2=5,
                                                                     multicomp_correction='holm',
                                                                     pval_thr=0.01,
                                                                     verbose=True)

    rel_sig_pairs = retrieve_relevant_from_nested_dict(computed_significance,
                                                       'stage2',
                                                       True,
                                                       allow_missing_keys=True)

    assert rel_sig_pairs == [(1, 49), (2, 48), (5, 45)] # retrieve correlated signals



from numba.experimental import jitclass
import numpy as np
from numba import int32, float32, boolean    # import the types
from numba import types, typed, njit


spec = [
    ('indices', types.ListType(types.float64)),
    ('ampls', types.ListType(types.float64)),
    ('birth_scale', float32),
    ('scales', types.ListType(types.float64)),
    ('wvt_times', types.ListType(types.float64)),
    ('terminated', boolean),
    ('end_scale', float32),
    ('length', float32),
    ('max_scale', float32),
    ('max_ampl', float32),
    ('start', float32),
    ('end', float32),
    ('duration', float32),
]


@njit()
def maxpos_numba(x):
    m = max(x)
    return x.index(m)


@jitclass(spec)
class Ridge(object):

    def __init__(self, start_index, ampl, start_scale, wvt_time):
        self.indices = typed.List.empty_list(types.float64)
        self.indices.append(start_index)

        self.ampls = typed.List.empty_list(types.float64)
        self.ampls.append(ampl)

        self.birth_scale = start_scale

        self.scales = typed.List.empty_list(types.float64)
        self.scales.append(start_scale)

        self.wvt_times = typed.List.empty_list(types.float64)
        self.wvt_times.append(wvt_time)

        self.terminated = False

        self.end_scale = -1
        self.length = -1
        self.max_scale = -1
        self.max_ampl = -1
        self.start = -1
        self.end = -1
        self.duration = -1


    def extend(self, index, ampl, scale, wvt_time):
        if not self.terminated:
            self.scales.append(scale)
            self.ampls.append(ampl)
            self.indices.append(index)
            self.wvt_times.append(wvt_time)
        else:
            raise ValueError('Ridge is terminated')


    def tip(self):
        return self.indices[-1]


    def terminate(self):
        if self.terminated:
            pass

        else:
            self.end_scale = self.scales[-1]
            self.length = len(self.scales)
            self.max_scale = self.scales[maxpos_numba(self.ampls)]
            self.max_ampl = max(self.ampls)
            self.start = self.indices[0]
            self.end = self.indices[-1]
            self.duration = np.abs(self.end-self.start)
            self.terminated = True


class RidgeInfoContainer(object):
    def __init__(self, indices, ampls, scales, wvt_times):
        self.indices = np.array(indices)
        self.ampls = np.array(ampls)
        self.scales = np.array(scales)
        self.wvt_times = np.array(wvt_times)

        self.birth_scale = scales[0]
        self.end_scale = scales[-1]
        self.length = len(self.scales)
        self.max_scale = self.scales[np.argmax(self.ampls)]
        self.max_ampl = max(self.ampls)
        self.start = self.indices[0]
        self.end = self.indices[-1]
        self.duration = np.abs(self.end - self.start)


def ridges_to_containers(ridges):
    rcs = [RidgeInfoContainer(ridge.indices, ridge.ampls, ridge.scales, ridge.wvt_times) for ridge in ridges]
    return rcs



import networkx as nx
import numpy as np
import scipy.sparse as sp

from src.driada.network.net_base import Network
from src.driada.network.graph_utils import get_giant_scc_from_graph
from src.driada.network.matrix_utils import turn_to_partially_directed


def create_default_adj(directed=False):
    seed = 42
    G = nx.random_regular_graph(5, 100, seed=seed)
    adj = nx.adjacency_matrix(G)
    adj = sp.csr_matrix(turn_to_partially_directed(adj.A, int(directed)))
    return adj


def create_default_graph(directed=False):
    adj = create_default_adj(directed=directed)
    gtype = nx.DiGraph if directed else nx.Graph
    graph = nx.from_scipy_sparse_array(adj, create_using=gtype)
    return graph


def create_default_net():
    adj = create_default_adj()
    net = Network(adj=adj)
    return net


def test_init_from_adj():
    adj = create_default_adj()
    net = Network(adj=adj, create_nx_graph=False)
    assert np.allclose(net.adj.data, adj.data)


def test_init_from_graph():
    g = create_default_graph()
    net = Network(graph=g)
    assert np.allclose(net.adj.data, nx.adjacency_matrix(g).data)
    assert nx.is_isomorphic(net.graph, g)


def test_remove_isolates():
    g = create_default_graph()
    g2 = g.copy()
    g2.add_nodes_from(['a', 'b', 'c'])

    net1 = Network(graph=g2, preprocessing='remove_isolates', verbose=True)
    assert np.allclose(net1.adj.data, nx.adjacency_matrix(g).data)
    assert nx.is_isomorphic(net1.graph, g)

    net2 = Network(adj=nx.adjacency_matrix(g2), create_nx_graph=False, preprocessing='remove_isolates')
    assert np.allclose(net2.adj.data, nx.adjacency_matrix(g).data)


def test_take_giant_cс():
    g = create_default_graph()
    g2 = g.copy()
    g2.add_nodes_from(['a', 'b', 'c', 'd', 'e'])
    g2.add_edges_from([('a', 'b'), ('c', 'd'), ('d', 'e'), ('e', 'c')])

    net1 = Network(graph=g2, preprocessing='giant_cc', verbose=True)
    assert np.allclose(net1.adj.data, nx.adjacency_matrix(g).data)
    assert nx.is_isomorphic(net1.graph, g)

    net2 = Network(adj=nx.adjacency_matrix(g2),
                   create_nx_graph=False,
                   preprocessing='giant_cc',
                   verbose=True)

    assert np.allclose(net2.adj.data, nx.adjacency_matrix(g).data)


def test_take_giant_scc():
    g = create_default_graph(directed=True)
    g2 = g.copy()
    g2.add_nodes_from(['a', 'b', 'c', 'd', 'e'])
    g2.add_edges_from([('a', 'b'), ('c', 'd'), ('d', 'e'), ('e', 'c')])

    gcc = get_giant_scc_from_graph(g2)

    net1 = Network(graph=g2, preprocessing='giant_scc', verbose=True)
    assert np.allclose(net1.adj.data, nx.adjacency_matrix(gcc).data)
    assert nx.is_isomorphic(net1.graph, gcc)

    net2 = Network(adj=nx.adjacency_matrix(g2),
                   create_nx_graph=False,
                   preprocessing='giant_scc',
                   verbose=True)

    assert np.allclose(net2.adj.data, nx.adjacency_matrix(gcc).data)



import numpy as np
from numba import njit
from scipy.stats import median_abs_deviation
from scipy.optimize import minimize
from ..signals.sig_base import TimeSeries
from .wavelet_event_detection import *

DEFAULT_T_RISE = 0.25 #sec
DEFAULT_T_OFF = 2.0 #sec

DEFAULT_FPS = 20.0 #frames per sec
DEFAULT_MIN_BEHAVIOUR_TIME = 0.25 #sec

MIN_CA_SHIFT = 5  # MIN_SHIFT*t_off is the minimal random signal shift for a given cell

#TODO: add numba decorators where possible
class Neuron():
    """
    Class for representing all information about a single neuron.

    Attributes
    ----------
    test: str
        description

    Methods
    -------
    test(arg=None)
        description
    """

    @staticmethod
    @njit()
    def spike_form(t, t_rise, t_off):
        form = (1-np.exp(-t/t_rise))*np.exp(-t/t_off)
        return form/max(form)


    @staticmethod
    def get_restored_calcium(sp, t_rise, t_off):
        x = np.linspace(0, 1000, num=1000)
        spform = Neuron.spike_form(x, t_rise=t_rise, t_off=t_off)
        conv = np.convolve(sp, spform)
        return conv

    @staticmethod
    def ca_mse_error(t_off, ca, spk, t_rise):
        # TODO: fix for new spike format
        re_ca = Neuron.get_restored_calcium(spk, t_rise, t_off)
        return np.sqrt(np.sum(np.abs(ca - re_ca[:len(ca)])**2)/len(ca))

    @staticmethod
    def calcium_preprocessing(ca):
        ca[np.where(ca < 0)[0]] = 0
        #ca = ca + np.abs(min(ca))
        ca += np.random.random(size=len(ca))*1e-8
        return ca

    def __init__(self, cell_id, ca, sp,
                 default_t_rise=DEFAULT_T_RISE, default_t_off=DEFAULT_T_OFF, fps=DEFAULT_FPS,
                 fit_individual_t_off=False):

        if default_t_rise is None:
            default_t_rise = DEFAULT_T_RISE
        if default_t_off is None:
            default_t_off = DEFAULT_T_OFF
        if fps is None:
            fps = DEFAULT_FPS

        self.cell_id = cell_id
        self.ca = TimeSeries(Neuron.calcium_preprocessing(ca), discrete=False)
        if sp is None:
            self.sp = None
        else:
            self.sp = TimeSeries(sp.astype(int), discrete=True)
        self.n_frames = len(ca.data)

        self.sp_count = np.sum(self.sp.data.astype(bool).astype(int))
        self.t_off = None
        self.noise_ampl = None
        self.mad = None
        self.snr = None

        self.default_t_off = default_t_off*fps
        self.default_t_rise = default_t_rise*fps

        if fit_individual_t_off:
            t_off = self.get_t_off()
        else:
            t_off = self.default_t_off

        # add shuffle mask according to computed characteristic calcium decay time
        self.ca.shuffle_mask = np.ones(self.n_frames).astype(bool)
        min_shift = int(t_off * MIN_CA_SHIFT)
        self.ca.shuffle_mask[:min_shift] = False
        self.ca.shuffle_mask[self.n_frames - min_shift:] = False

    def reconstruct_spikes(self, **kwargs):
        raise AttributeError('Spike reconstruction not implemented')

    def get_mad(self):
        if self.mad is None:
            try:
                self.snr, self.mad = self._calc_snr()
            except ValueError:
                self.mad = median_abs_deviation(self.ca.data)
        return self.mad

    def get_snr(self):
        if self.snr is None:
            self.snr, self.mad = self._calc_snr()
        return self.snr

    def _calc_snr(self):
        spk_inds = np.nonzero(self.sp.data)[0]
        mad = median_abs_deviation(self.ca.data)
        if len(spk_inds) > 0:
            sn = np.mean(self.ca.data[spk_inds])/mad
            if np.isnan(sn):
                raise ValueError('Error in snr calculation')
        else:
            raise ValueError('No spikes found!')

        return sn, mad

    def get_t_off(self):
        if self.t_off is None:
            self.t_off, self.noise_ampl = self._fit_t_off()

        return self.t_off

    def get_noise_ampl(self):
        if self.noise_ampl is None:
            self.t_off, self.noise_ampl = self._fit_t_off()

        return self.noise_ampl


    def _fit_t_off(self):

        #TODO: fit for an arbitrary kernel form.
        #TODO: add nonlinear summation fit if needed

        res = minimize(Neuron.ca_mse_error, (np.array([self.default_t_off])), args=(self.ca.data, self.sp.data, self.default_t_rise))
        opt_t_off = res.x[0]
        noise_amplitude = res.fun

        if opt_t_off > self.default_t_off*5:
            print(f'Calculated t_off={int(opt_t_off)} for neuron {self.cell_id} is suspiciously high, check signal quality. t_off has been automatically lowered to {self.default_t_off*5}')

        return min(opt_t_off, self.default_t_off*5), noise_amplitude


    def get_shuffled_calcium(self, method = 'roll_based', no_ts=True, **kwargs):
        try:
            fn = getattr(self, f'_shuffle_calcium_data_{method}')
        except AttributeError():
            raise UserWarning('Unknown calcium data shuffling method')

        sh_ca = fn(**kwargs)
        sh_ca = Neuron.calcium_preprocessing(sh_ca)
        if not no_ts:
            sh_ca = TimeSeries(Neuron.calcium_preprocessing(sh_ca), discrete=False)

        return sh_ca


    def _shuffle_calcium_data_waveform_based(self, **kwargs):

        shuf_ca = np.zeros(self.n_frames)
        opt_t_off, noise_amplitude = self.get_t_off(), self.get_noise_ampl()

        #noise = np.random.normal(loc = 0, scale = noise_amplitude, size = len(self.ca))

        conv = Neuron.get_restored_calcium(self.sp.data, 5, opt_t_off)
        background = self.ca.data - conv[:len(self.ca.data)]

        pspk = self._shuffle_spikes_data_isi_based()
        psconv = Neuron.get_restored_calcium(pspk, 5, opt_t_off)

        #shuf_ca = conv[:len(self.ca.data)] + noise
        shuf_ca = psconv[:len(self.ca.data)] + background
        return shuf_ca


    def _shuffle_calcium_data_chunks_based(self, **kwargs):
        if 'n' not in kwargs:
            n = 100
        else:
            n = kwargs['n']

        shuf_ca = np.zeros(self.n_frames)
        ca = self.ca.data
        chunks = np.concatenate(np.split(ca[:-len(ca)%n], n), ca[-(len(ca)%n):])
        inds = np.arange(n)
        np.random.shuffle(inds)

        shuf_ca[:] = np.concatenate(tuple(np.array(chunks)[inds]))

        return shuf_ca


    def _shuffle_calcium_data_roll_based(self, **kwargs):
        opt_t_off = self.get_t_off()
        if 'shift' in kwargs:
            shift = kwargs['shift']
        else:
            shift = np.random.randint(3*opt_t_off, self.n_frames - 3*opt_t_off)

        shuf_ca = np.roll(self.ca.data, shift)

        return shuf_ca


    def get_shuffled_spikes(self, method = 'isi_based', no_ts=True, **kwargs):
        if self.sp is None:
            raise AttributeError('Unable to shuffle spikes without spikes data')

        try:
            fn = getattr(self, f'_shuffle_spikes_data_{method}')
        except AttributeError():
            raise UserWarning('Unknown calcium data shuffling method')

        sh_data = fn(**kwargs)
        if not no_ts:
            return TimeSeries(sh_data, discrete=True)
        else:
            return sh_data


    def _shuffle_spikes_data_isi_based(self):
        nfr = self.n_frames

        pseudo_spikes = np.zeros(nfr)
        event_inds = np.where(self.sp.data != 0)[0]

        if len(event_inds) == 0: #if no events were detected, there is nothing to shuffle
            return self.sp.data

        event_vals = self.sp.data[event_inds]
        first_random_pos = np.random.choice(nfr - (max(event_inds) - min(event_inds)))

        interspike_intervals = np.diff(event_inds)
        rng = np.arange(len(interspike_intervals))
        np.random.shuffle(rng)
        disordered_interspike_intervals = interspike_intervals[rng]

        pseudo_event_inds = np.cumsum(np.insert(disordered_interspike_intervals,
                                                0, first_random_pos))

        pseudo_event_vals = event_vals
        np.random.shuffle(event_vals)
        pseudo_spikes[pseudo_event_inds] = pseudo_event_vals

        return pseudo_spikes
